# This file is part of Buildbot.  Buildbot is free software: you can
# redistribute it and/or modify it under the terms of the GNU General Public
# License as published by the Free Software Foundation, version 2.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
# details.
#
# You should have received a copy of the GNU General Public License along with
# this program; if not, write to the Free Software Foundation, Inc., 51
# Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#
# Copyright Buildbot Team Members

import mock

from twisted.internet import defer
from twisted.trial import unittest

from buildbot.plugins import util
from buildbot.secrets.manager import SecretManager
from buildbot.test.fake.secrets import FakeSecretStorage
from buildbot.test.fake.web import FakeRequest
from buildbot.test.fake.web import fakeMasterForHooks
from buildbot.test.util.misc import TestReactorMixin
from buildbot.www import change_hook
from buildbot.www.hooks.gitlab import _HEADER_EVENT
from buildbot.www.hooks.gitlab import _HEADER_GITLAB_TOKEN

# Sample GITLAB commit payload from https://docs.gitlab.com/ce/user/project/integrations/webhooks.html  # noqa pylint: disable=line-too-long
# Added "modified" and "removed", and change email
gitJsonPayload = b"""
{
  "before": "95790bf891e76fee5e1747ab589903a6a1f80f22",
  "after": "da1560886d4f094c3e6c9ef40349f7d38b5d27d7",
  "ref": "refs/heads/master",
  "user_id": 4,
  "user_name": "John Smith",
  "repository": {
    "name": "Diaspora",
    "url": "git@localhost:diaspora.git",
    "description": "",
    "homepage": "http://localhost/diaspora"
  },
  "commits": [
    {
      "id": "b6568db1bc1dcd7f8b4d5a946b0b91f9dacd7327",
      "message": "Update Catalan translation to e38cb41.",
      "timestamp": "2011-12-12T14:27:31+02:00",
      "url": "http://localhost/diaspora/commits/b6568db1bc1dcd7f8b4d5a946b0b91f9dacd7327",
      "author": {
        "name": "Jordi Mallach",
        "email": "jordi@softcatala.org"
      }
    },
    {
      "id": "da1560886d4f094c3e6c9ef40349f7d38b5d27d7",
      "message": "fixed readme",
      "timestamp": "2012-01-03T23:36:29+02:00",
      "url": "http://localhost/diaspora/commits/da1560886d4f094c3e6c9ef40349f7d38b5d27d7",
      "author": {
        "name": "GitLab dev user",
        "email": "gitlabdev@dv6700.(none)"
      }
    }
  ],
  "total_commits_count": 2
}
"""
gitJsonPayloadTag = b"""
{
  "object_kind": "tag_push",
  "before": "0000000000000000000000000000000000000000",
  "after": "82b3d5ae55f7080f1e6022629cdb57bfae7cccc7",
  "ref": "refs/tags/v1.0.0",
  "checkout_sha": "82b3d5ae55f7080f1e6022629cdb57bfae7cccc7",
  "user_id": 1,
  "user_name": "John Smith",
  "repository":{
    "name": "Example",
    "url": "git@localhost:diaspora.git",
    "description": "",
    "homepage": "http://example.com/jsmith/example",
    "git_http_url":"http://example.com/jsmith/example.git",
    "git_ssh_url":"git@example.com:jsmith/example.git",
    "visibility_level":0
  },
   "commits": [
     {
       "id": "b6568db1bc1dcd7f8b4d5a946b0b91f9dacd7327",
       "message": "Update Catalan translation to e38cb41.",
       "timestamp": "2011-12-12T14:27:31+02:00",
       "url": "http://localhost/diaspora/commits/b6568db1bc1dcd7f8b4d5a946b0b91f9dacd7327",
       "author": {
         "name": "Jordi Mallach",
         "email": "jordi@softcatala.org"
       }
     },
     {
       "id": "da1560886d4f094c3e6c9ef40349f7d38b5d27d7",
       "message": "fixed readme",
       "timestamp": "2012-01-03T23:36:29+02:00",
       "url": "http://localhost/diaspora/commits/da1560886d4f094c3e6c9ef40349f7d38b5d27d7",
       "author": {
         "name": "GitLab dev user",
         "email": "gitlabdev@dv6700.(none)"
       }
     }
   ],
   "total_commits_count": 2
}
"""

# == Merge requests from a different branch of the same project
# GITLAB commit payload from an actual version 10.7.1-ee gitlab instance
# chronicling the lives and times of a trivial MR through the operations
# open, edit description, add commit, close, and reopen, in that order.
# (Tidied with json_pp --json_opt=canonical,pretty and an editor.)
# FIXME: only show diffs here to keep file smaller and increase clarity
gitJsonPayloadMR_open = b"""
{
   "event_type" : "merge_request",
   "object_attributes" : {
      "action" : "open",
      "assignee_id" : null,
      "author_id" : 15,
      "created_at" : "2018-05-15 07:45:37 -0700",
      "description" : "This to both gitlab gateways!",
      "head_pipeline_id" : 29931,
      "human_time_estimate" : null,
      "human_total_time_spent" : null,
      "id" : 10850,
      "iid" : 6,
      "last_commit" : {
         "author" : {
            "email" : "mmusterman@example.com",
            "name" : "Max Mustermann"
         },
         "id" : "92268bc781b24f0a61b907da062950e9e5252a69",
         "message" : "Remove the dummy line again",
         "timestamp" : "2018-05-14T07:54:04-07:00",
         "url" : "https://gitlab.example.com/mmusterman/awesome_project/commit/92268bc781b24f0a61b907da062950e9e5252a69"
      },
      "last_edited_at" : null,
      "last_edited_by_id" : null,
      "merge_commit_sha" : null,
      "merge_error" : null,
      "merge_params" : {
         "force_remove_source_branch" : 0
      },
      "merge_status" : "unchecked",
      "merge_user_id" : null,
      "merge_when_pipeline_succeeds" : false,
      "milestone_id" : null,
      "source" : {
         "avatar_url" : null,
         "ci_config_path" : null,
         "default_branch" : "master",
         "description" : "Trivial project for testing build machinery quickly",
         "git_http_url" : "https://gitlab.example.com/mmusterman/awesome_project.git",
         "git_ssh_url" : "git@gitlab.example.com:mmusterman/awesome_project.git",
         "homepage" : "https://gitlab.example.com/mmusterman/awesome_project",
         "http_url" : "https://gitlab.example.com/mmusterman/awesome_project.git",
         "id" : 239,
         "name" : "awesome_project",
         "namespace" : "mmusterman",
         "path_with_namespace" : "mmusterman/awesome_project",
         "ssh_url" : "git@gitlab.example.com:mmusterman/awesome_project.git",
         "url" : "git@gitlab.example.com:mmusterman/awesome_project.git",
         "visibility_level" : 0,
         "web_url" : "https://gitlab.example.com/mmusterman/awesome_project"
      },
      "source_branch" : "ms-viewport",
      "source_project_id" : 239,
      "state" : "opened",
      "target" : {
         "avatar_url" : null,
         "ci_config_path" : null,
         "default_branch" : "master",
         "description" : "Trivial project for testing build machinery quickly",
         "git_http_url" : "https://gitlab.example.com/mmusterman/awesome_project.git",
         "git_ssh_url" : "git@gitlab.example.com:mmusterman/awesome_project.git",
         "homepage" : "https://gitlab.example.com/mmusterman/awesome_project",
         "http_url" : "https://gitlab.example.com/mmusterman/awesome_project.git",
         "id" : 239,
         "name" : "awesome_project",
         "namespace" : "mmusterman",
         "path_with_namespace" : "mmusterman/awesome_project",
         "ssh_url" : "git@gitlab.example.com:mmusterman/awesome_project.git",
         "url" : "git@gitlab.example.com:mmusterman/awesome_project.git",
         "visibility_level" : 0,
         "web_url" : "https://gitlab.example.com/mmusterman/awesome_project"
      },
      "target_branch" : "master",
      "target_project_id" : 239,
      "time_estimate" : 0,
      "title" : "Remove the dummy line again",
      "total_time_spent" : 0,
      "updated_at" : "2018-05-15 07:45:37 -0700",
      "updated_by_id" : null,
      "url" : "https://gitlab.example.com/mmusterman/awesome_project/merge_requests/6",
      "work_in_progress" : false
   },
   "object_kind" : "merge_request",
   "project" : {
      "avatar_url" : null,
      "ci_config_path" : null,
      "default_branch" : "master",
      "description" : "Trivial project for testing build machinery quickly",
      "git_http_url" : "https://gitlab.example.com/mmusterman/awesome_project.git",
      "git_ssh_url" : "git@gitlab.example.com:mmusterman/awesome_project.git",
      "homepage" : "https://gitlab.example.com/mmusterman/awesome_project",
      "http_url" : "https://gitlab.example.com/mmusterman/awesome_project.git",
      "id" : 239,
      "name" : "awesome_project",
      "namespace" : "mmusterman",
      "path_with_namespace" : "mmusterman/awesome_project",
      "ssh_url" : "git@gitlab.example.com:mmusterman/awesome_project.git",
      "url" : "git@gitlab.example.com:mmusterman/awesome_project.git",
      "visibility_level" : 0,
      "web_url" : "https://gitlab.example.com/mmusterman/awesome_project"
   },
   "user" : {
      "avatar_url" : "http://www.gravatar.com/avatar/e64c7d89f26bd1972efa854d13d7dd61?s=40&d=identicon",
      "name" : "Max Mustermann",
      "username" : "mmusterman"
   }
}
"""  # noqa pylint: disable=line-too-long
gitJsonPayloadMR_editdesc = b"""
{
   "event_type" : "merge_request",
   "object_attributes" : {
      "action" : "update",
      "assignee_id" : null,
      "author_id" : 15,
      "created_at" : "2018-05-15 07:45:37 -0700",
      "description" : "Edited description.",
      "head_pipeline_id" : 29931,
      "human_time_estimate" : null,
      "human_total_time_spent" : null,
      "id" : 10850,
      "iid" : 6,
      "last_commit" : {
         "author" : {
            "email" : "mmusterman@example.com",
            "name" : "Max Mustermann"
         },
         "id" : "92268bc781b24f0a61b907da062950e9e5252a69",
         "message" : "Remove the dummy line again",
         "timestamp" : "2018-05-14T07:54:04-07:00",
         "url" : "https://gitlab.example.com/mmusterman/awesome_project/commit/92268bc781b24f0a61b907da062950e9e5252a69"
      },
      "last_edited_at" : "2018-05-15 07:49:55 -0700",
      "last_edited_by_id" : 15,
      "merge_commit_sha" : null,
      "merge_error" : null,
      "merge_params" : {
         "force_remove_source_branch" : 0
      },
      "merge_status" : "can_be_merged",
      "merge_user_id" : null,
      "merge_when_pipeline_succeeds" : false,
      "milestone_id" : null,
      "source" : {
         "avatar_url" : null,
         "ci_config_path" : null,
         "default_branch" : "master",
         "description" : "Trivial project for testing build machinery quickly",
         "git_http_url" : "https://gitlab.example.com/mmusterman/awesome_project.git",
         "git_ssh_url" : "git@gitlab.example.com:mmusterman/awesome_project.git",
         "homepage" : "https://gitlab.example.com/mmusterman/awesome_project",
         "http_url" : "https://gitlab.example.com/mmusterman/awesome_project.git",
         "id" : 239,
         "name" : "awesome_project",
         "namespace" : "mmusterman",
         "path_with_namespace" : "mmusterman/awesome_project",
         "ssh_url" : "git@gitlab.example.com:mmusterman/awesome_project.git",
         "url" : "git@gitlab.example.com:mmusterman/awesome_project.git",
         "visibility_level" : 0,
         "web_url" : "https://gitlab.example.com/mmusterman/awesome_project"
      },
      "source_branch" : "ms-viewport",
      "source_project_id" : 239,
      "state" : "opened",
      "target" : {
         "avatar_url" : null,
         "ci_config_path" : null,
         "default_branch" : "master",
         "description" : "Trivial project for testing build machinery quickly",
         "git_http_url" : "https://gitlab.example.com/mmusterman/awesome_project.git",
         "git_ssh_url" : "git@gitlab.example.com:mmusterman/awesome_project.git",
         "homepage" : "https://gitlab.example.com/mmusterman/awesome_project",
         "http_url" : "https://gitlab.example.com/mmusterman/awesome_project.git",
         "id" : 239,
         "name" : "awesome_project",
         "namespace" : "mmusterman",
         "path_with_namespace" : "mmusterman/awesome_project",
         "ssh_url" : "git@gitlab.example.com:mmusterman/awesome_project.git",
         "url" : "git@gitlab.example.com:mmusterman/awesome_project.git",
         "visibility_level" : 0,
         "web_url" : "https://gitlab.example.com/mmusterman/awesome_project"
      },
      "target_branch" : "master",
      "target_project_id" : 239,
      "time_estimate" : 0,
      "title" : "Remove the dummy line again",
      "total_time_spent" : 0,
      "updated_at" : "2018-05-15 07:49:55 -0700",
      "updated_by_id" : 15,
      "url" : "https://gitlab.example.com/mmusterman/awesome_project/merge_requests/6",
      "work_in_progress" : false
   },
   "object_kind" : "merge_request",
   "project" : {
      "avatar_url" : null,
      "ci_config_path" : null,
      "default_branch" : "master",
      "description" : "Trivial project for testing build machinery quickly",
      "git_http_url" : "https://gitlab.example.com/mmusterman/awesome_project.git",
      "git_ssh_url" : "git@gitlab.example.com:mmusterman/awesome_project.git",
      "homepage" : "https://gitlab.example.com/mmusterman/awesome_project",
      "http_url" : "https://gitlab.example.com/mmusterman/awesome_project.git",
      "id" : 239,
      "name" : "awesome_project",
      "namespace" : "mmusterman",
      "path_with_namespace" : "mmusterman/awesome_project",
      "ssh_url" : "git@gitlab.example.com:mmusterman/awesome_project.git",
      "url" : "git@gitlab.example.com:mmusterman/awesome_project.git",
      "visibility_level" : 0,
      "web_url" : "https://gitlab.example.com/mmusterman/awesome_project"
   },
   "user" : {
      "avatar_url" : "http://www.gravatar.com/avatar/e64c7d89f26bd1972efa854d13d7dd61?s=40&d=identicon",
      "name" : "Max Mustermann",
      "username" : "mmusterman"
   }
}
"""  # noqa pylint: disable=line-too-long
gitJsonPayloadMR_addcommit = b"""
{
   "event_type" : "merge_request",
   "object_attributes" : {
      "action" : "update",
      "assignee_id" : null,
      "author_id" : 15,
      "created_at" : "2018-05-15 07:45:37 -0700",
      "description" : "Edited description.",
      "head_pipeline_id" : 29931,
      "human_time_estimate" : null,
      "human_total_time_spent" : null,
      "id" : 10850,
      "iid" : 6,
      "last_commit" : {
         "author" : {
            "email" : "mmusterman@example.com",
            "name" : "Max Mustermann"
         },
         "id" : "cee8b01dcbaeed89563c2822f7c59a93c813eb6b",
         "message" : "debian/compat: update to 9",
         "timestamp" : "2018-05-15T07:51:11-07:00",
         "url" : "https://gitlab.example.com/mmusterman/awesome_project/commit/cee8b01dcbaeed89563c2822f7c59a93c813eb6b"
      },
      "last_edited_at" : "2018-05-15 14:49:55 UTC",
      "last_edited_by_id" : 15,
      "merge_commit_sha" : null,
      "merge_error" : null,
      "merge_params" : {
         "force_remove_source_branch" : 0
      },
      "merge_status" : "unchecked",
      "merge_user_id" : null,
      "merge_when_pipeline_succeeds" : false,
      "milestone_id" : null,
      "oldrev" : "92268bc781b24f0a61b907da062950e9e5252a69",
      "source" : {
         "avatar_url" : null,
         "ci_config_path" : null,
         "default_branch" : "master",
         "description" : "Trivial project for testing build machinery quickly",
         "git_http_url" : "https://gitlab.example.com/mmusterman/awesome_project.git",
         "git_ssh_url" : "git@gitlab.example.com:mmusterman/awesome_project.git",
         "homepage" : "https://gitlab.example.com/mmusterman/awesome_project",
         "http_url" : "https://gitlab.example.com/mmusterman/awesome_project.git",
         "id" : 239,
         "name" : "awesome_project",
         "namespace" : "mmusterman",
         "path_with_namespace" : "mmusterman/awesome_project",
         "ssh_url" : "git@gitlab.example.com:mmusterman/awesome_project.git",
         "url" : "git@gitlab.example.com:mmusterman/awesome_project.git",
         "visibility_level" : 0,
         "web_url" : "https://gitlab.example.com/mmusterman/awesome_project"
      },
      "source_branch" : "ms-viewport",
      "source_project_id" : 239,
      "state" : "opened",
      "target" : {
         "avatar_url" : null,
         "ci_config_path" : null,
         "default_branch" : "master",
         "description" : "Trivial project for testing build machinery quickly",
         "git_http_url" : "https://gitlab.example.com/mmusterman/awesome_project.git",
         "git_ssh_url" : "git@gitlab.example.com:mmusterman/awesome_project.git",
         "homepage" : "https://gitlab.example.com/mmusterman/awesome_project",
         "http_url" : "https://gitlab.example.com/mmusterman/awesome_project.git",
         "id" : 239,
         "name" : "awesome_project",
         "namespace" : "mmusterman",
         "path_with_namespace" : "mmusterman/awesome_project",
         "ssh_url" : "git@gitlab.example.com:mmusterman/awesome_project.git",
         "url" : "git@gitlab.example.com:mmusterman/awesome_project.git",
         "visibility_level" : 0,
         "web_url" : "https://gitlab.example.com/mmusterman/awesome_project"
      },
      "target_branch" : "master",
      "target_project_id" : 239,
      "time_estimate" : 0,
      "title" : "Remove the dummy line again",
      "total_time_spent" : 0,
      "updated_at" : "2018-05-15 14:51:27 UTC",
      "updated_by_id" : 15,
      "url" : "https://gitlab.example.com/mmusterman/awesome_project/merge_requests/6",
      "work_in_progress" : false
   },
   "object_kind" : "merge_request",
   "project" : {
      "avatar_url" : null,
      "ci_config_path" : null,
      "default_branch" : "master",
      "description" : "Trivial project for testing build machinery quickly",
      "git_http_url" : "https://gitlab.example.com/mmusterman/awesome_project.git",
      "git_ssh_url" : "git@gitlab.example.com:mmusterman/awesome_project.git",
      "homepage" : "https://gitlab.example.com/mmusterman/awesome_project",
      "http_url" : "https://gitlab.example.com/mmusterman/awesome_project.git",
      "id" : 239,
      "name" : "awesome_project",
      "namespace" : "mmusterman",
      "path_with_namespace" : "mmusterman/awesome_project",
      "ssh_url" : "git@gitlab.example.com:mmusterman/awesome_project.git",
      "url" : "git@gitlab.example.com:mmusterman/awesome_project.git",
      "visibility_level" : 0,
      "web_url" : "https://gitlab.example.com/mmusterman/awesome_project"
   },
   "user" : {
      "avatar_url" : "http://www.gravatar.com/avatar/e64c7d89f26bd1972efa854d13d7dd61?s=40&d=identicon",
      "name" : "Max Mustermann",
      "username" : "mmusterman"
   }
}
"""  # noqa pylint: disable=line-too-long
gitJsonPayloadMR_close = b"""
{
   "event_type" : "merge_request",
   "object_attributes" : {
      "action" : "close",
      "assignee_id" : null,
      "author_id" : 15,
      "created_at" : "2018-05-15 07:45:37 -0700",
      "description" : "Edited description.",
      "head_pipeline_id" : 29958,
      "human_time_estimate" : null,
      "human_total_time_spent" : null,
      "id" : 10850,
      "iid" : 6,
      "last_commit" : {
         "author" : {
            "email" : "mmusterman@example.com",
            "name" : "Max Mustermann"
         },
         "id" : "cee8b01dcbaeed89563c2822f7c59a93c813eb6b",
         "message" : "debian/compat: update to 9",
         "timestamp" : "2018-05-15T07:51:11-07:00",
         "url" : "https://gitlab.example.com/mmusterman/awesome_project/commit/cee8b01dcbaeed89563c2822f7c59a93c813eb6b"
      },
      "last_edited_at" : "2018-05-15 07:49:55 -0700",
      "last_edited_by_id" : 15,
      "merge_commit_sha" : null,
      "merge_error" : null,
      "merge_params" : {
         "force_remove_source_branch" : 0
      },
      "merge_status" : "can_be_merged",
      "merge_user_id" : null,
      "merge_when_pipeline_succeeds" : false,
      "milestone_id" : null,
      "source" : {
         "avatar_url" : null,
         "ci_config_path" : null,
         "default_branch" : "master",
         "description" : "Trivial project for testing build machinery quickly",
         "git_http_url" : "https://gitlab.example.com/mmusterman/awesome_project.git",
         "git_ssh_url" : "git@gitlab.example.com:mmusterman/awesome_project.git",
         "homepage" : "https://gitlab.example.com/mmusterman/awesome_project",
         "http_url" : "https://gitlab.example.com/mmusterman/awesome_project.git",
         "id" : 239,
         "name" : "awesome_project",
         "namespace" : "mmusterman",
         "path_with_namespace" : "mmusterman/awesome_project",
         "ssh_url" : "git@gitlab.example.com:mmusterman/awesome_project.git",
         "url" : "git@gitlab.example.com:mmusterman/awesome_project.git",
         "visibility_level" : 0,
         "web_url" : "https://gitlab.example.com/mmusterman/awesome_project"
      },
      "source_branch" : "ms-viewport",
      "source_project_id" : 239,
      "state" : "closed",
      "target" : {
         "avatar_url" : null,
         "ci_config_path" : null,
         "default_branch" : "master",
         "description" : "Trivial project for testing build machinery quickly",
         "git_http_url" : "https://gitlab.example.com/mmusterman/awesome_project.git",
         "git_ssh_url" : "git@gitlab.example.com:mmusterman/awesome_project.git",
         "homepage" : "https://gitlab.example.com/mmusterman/awesome_project",
         "http_url" : "https://gitlab.example.com/mmusterman/awesome_project.git",
         "id" : 239,
         "name" : "awesome_project",
         "namespace" : "mmusterman",
         "path_with_namespace" : "mmusterman/awesome_project",
         "ssh_url" : "git@gitlab.example.com:mmusterman/awesome_project.git",
         "url" : "git@gitlab.example.com:mmusterman/awesome_project.git",
         "visibility_level" : 0,
         "web_url" : "https://gitlab.example.com/mmusterman/awesome_project"
      },
      "target_branch" : "master",
      "target_project_id" : 239,
      "time_estimate" : 0,
      "title" : "Remove the dummy line again",
      "total_time_spent" : 0,
      "updated_at" : "2018-05-15 07:52:01 -0700",
      "updated_by_id" : 15,
      "url" : "https://gitlab.example.com/mmusterman/awesome_project/merge_requests/6",
      "work_in_progress" : false
   },
   "object_kind" : "merge_request",
   "project" : {
      "avatar_url" : null,
      "ci_config_path" : null,
      "default_branch" : "master",
      "description" : "Trivial project for testing build machinery quickly",
      "git_http_url" : "https://gitlab.example.com/mmusterman/awesome_project.git",
      "git_ssh_url" : "git@gitlab.example.com:mmusterman/awesome_project.git",
      "homepage" : "https://gitlab.example.com/mmusterman/awesome_project",
      "http_url" : "https://gitlab.example.com/mmusterman/awesome_project.git",
      "id" : 239,
      "name" : "awesome_project",
      "namespace" : "mmusterman",
      "path_with_namespace" : "mmusterman/awesome_project",
      "ssh_url" : "git@gitlab.example.com:mmusterman/awesome_project.git",
      "url" : "git@gitlab.example.com:mmusterman/awesome_project.git",
      "visibility_level" : 0,
      "web_url" : "https://gitlab.example.com/mmusterman/awesome_project"
   },
   "user" : {
      "avatar_url" : "http://www.gravatar.com/avatar/e64c7d89f26bd1972efa854d13d7dd61?s=40&d=identicon",
      "name" : "Max Mustermann",
      "username" : "mmusterman"
   }
}
"""  # noqa pylint: disable=line-too-long
gitJsonPayloadMR_reopen = b"""
{
   "event_type" : "merge_request",
   "object_attributes" : {
      "action" : "reopen",
      "assignee_id" : null,
      "author_id" : 15,
      "created_at" : "2018-05-15 07:45:37 -0700",
      "description" : "Edited description.",
      "head_pipeline_id" : 29958,
      "human_time_estimate" : null,
      "human_total_time_spent" : null,
      "id" : 10850,
      "iid" : 6,
      "last_commit" : {
         "author" : {
            "email" : "mmusterman@example.com",
            "name" : "Max Mustermann"
         },
         "id" : "cee8b01dcbaeed89563c2822f7c59a93c813eb6b",
         "message" : "debian/compat: update to 9",
         "timestamp" : "2018-05-15T07:51:11-07:00",
         "url" : "https://gitlab.example.com/mmusterman/awesome_project/commit/cee8b01dcbaeed89563c2822f7c59a93c813eb6b"
      },
      "last_edited_at" : "2018-05-15 07:49:55 -0700",
      "last_edited_by_id" : 15,
      "merge_commit_sha" : null,
      "merge_error" : null,
      "merge_params" : {
         "force_remove_source_branch" : 0
      },
      "merge_status" : "can_be_merged",
      "merge_user_id" : null,
      "merge_when_pipeline_succeeds" : false,
      "milestone_id" : null,
      "source" : {
         "avatar_url" : null,
         "ci_config_path" : null,
         "default_branch" : "master",
         "description" : "Trivial project for testing build machinery quickly",
         "git_http_url" : "https://gitlab.example.com/mmusterman/awesome_project.git",
         "git_ssh_url" : "git@gitlab.example.com:mmusterman/awesome_project.git",
         "homepage" : "https://gitlab.example.com/mmusterman/awesome_project",
         "http_url" : "https://gitlab.example.com/mmusterman/awesome_project.git",
         "id" : 239,
         "name" : "awesome_project",
         "namespace" : "mmusterman",
         "path_with_namespace" : "mmusterman/awesome_project",
         "ssh_url" : "git@gitlab.example.com:mmusterman/awesome_project.git",
         "url" : "git@gitlab.example.com:mmusterman/awesome_project.git",
         "visibility_level" : 0,
         "web_url" : "https://gitlab.example.com/mmusterman/awesome_project"
      },
      "source_branch" : "ms-viewport",
      "source_project_id" : 239,
      "state" : "opened",
      "target" : {
         "avatar_url" : null,
         "ci_config_path" : null,
         "default_branch" : "master",
         "description" : "Trivial project for testing build machinery quickly",
         "git_http_url" : "https://gitlab.example.com/mmusterman/awesome_project.git",
         "git_ssh_url" : "git@gitlab.example.com:mmusterman/awesome_project.git",
         "homepage" : "https://gitlab.example.com/mmusterman/awesome_project",
         "http_url" : "https://gitlab.example.com/mmusterman/awesome_project.git",
         "id" : 239,
         "name" : "awesome_project",
         "namespace" : "mmusterman",
         "path_with_namespace" : "mmusterman/awesome_project",
         "ssh_url" : "git@gitlab.example.com:mmusterman/awesome_project.git",
         "url" : "git@gitlab.example.com:mmusterman/awesome_project.git",
         "visibility_level" : 0,
         "web_url" : "https://gitlab.example.com/mmusterman/awesome_project"
      },
      "target_branch" : "master",
      "target_project_id" : 239,
      "time_estimate" : 0,
      "title" : "Remove the dummy line again",
      "total_time_spent" : 0,
      "updated_at" : "2018-05-15 07:53:27 -0700",
      "updated_by_id" : 15,
      "url" : "https://gitlab.example.com/mmusterman/awesome_project/merge_requests/6",
      "work_in_progress" : false
   },
   "object_kind" : "merge_request",
   "project" : {
      "avatar_url" : null,
      "ci_config_path" : null,
      "default_branch" : "master",
      "description" : "Trivial project for testing build machinery quickly",
      "git_http_url" : "https://gitlab.example.com/mmusterman/awesome_project.git",
      "git_ssh_url" : "git@gitlab.example.com:mmusterman/awesome_project.git",
      "homepage" : "https://gitlab.example.com/mmusterman/awesome_project",
      "http_url" : "https://gitlab.example.com/mmusterman/awesome_project.git",
      "id" : 239,
      "name" : "awesome_project",
      "namespace" : "mmusterman",
      "path_with_namespace" : "mmusterman/awesome_project",
      "ssh_url" : "git@gitlab.example.com:mmusterman/awesome_project.git",
      "url" : "git@gitlab.example.com:mmusterman/awesome_project.git",
      "visibility_level" : 0,
      "web_url" : "https://gitlab.example.com/mmusterman/awesome_project"
   },
   "user" : {
      "avatar_url" : "http://www.gravatar.com/avatar/e64c7d89f26bd1972efa854d13d7dd61?s=40&d=identicon",
      "name" : "Max Mustermann",
      "username" : "mmusterman"
   }
}
"""  # noqa pylint: disable=line-too-long

# == Merge requests from a fork of the project
# (Captured more accurately than above test data)
gitJsonPayloadMR_open_forked = b"""
{
   "changes" : {
      "total_time_spent" : {
         "current" : 0,
         "previous" : null
      }
   },
   "event_type" : "merge_request",
   "labels" : [],
   "object_attributes" : {
      "action" : "open",
      "assignee_id" : null,
      "author_id" : 15,
      "created_at" : "2018-05-19 06:57:12 -0700",
      "description" : "This is a merge request from a fork of the project.",
      "head_pipeline_id" : null,
      "human_time_estimate" : null,
      "human_total_time_spent" : null,
      "id" : 10914,
      "iid" : 7,
      "last_commit" : {
         "author" : {
            "email" : "mmusterman@example.com",
            "name" : "Max Mustermann"
         },
         "id" : "e46ee239f3d6d41ade4d1e610669dd71ed86ec80",
         "message" : "Add note to README",
         "timestamp" : "2018-05-19T06:35:26-07:00",
         "url" : "https://gitlab.example.com/mmusterman/awesome_project/commit/e46ee239f3d6d41ade4d1e610669dd71ed86ec80"
      },
      "last_edited_at" : null,
      "last_edited_by_id" : null,
      "merge_commit_sha" : null,
      "merge_error" : null,
      "merge_params" : {
         "force_remove_source_branch" : "0"
      },
      "merge_status" : "unchecked",
      "merge_user_id" : null,
      "merge_when_pipeline_succeeds" : false,
      "milestone_id" : null,
      "source" : {
         "avatar_url" : null,
         "ci_config_path" : null,
         "default_branch" : "master",
         "description" : "Trivial project for testing build machinery quickly",
         "git_http_url" : "https://gitlab.example.com/build/awesome_project.git",
         "git_ssh_url" : "git@gitlab.example.com:build/awesome_project.git",
         "homepage" : "https://gitlab.example.com/build/awesome_project",
         "http_url" : "https://gitlab.example.com/build/awesome_project.git",
         "id" : 2337,
         "name" : "awesome_project",
         "namespace" : "build",
         "path_with_namespace" : "build/awesome_project",
         "ssh_url" : "git@gitlab.example.com:build/awesome_project.git",
         "url" : "git@gitlab.example.com:build/awesome_project.git",
         "visibility_level" : 0,
         "web_url" : "https://gitlab.example.com/build/awesome_project"
      },
      "source_branch" : "ms-viewport",
      "source_project_id" : 2337,
      "state" : "opened",
      "target" : {
         "avatar_url" : null,
         "ci_config_path" : null,
         "default_branch" : "master",
         "description" : "Trivial project for testing build machinery quickly",
         "git_http_url" : "https://gitlab.example.com/mmusterman/awesome_project.git",
         "git_ssh_url" : "git@gitlab.example.com:mmusterman/awesome_project.git",
         "homepage" : "https://gitlab.example.com/mmusterman/awesome_project",
         "http_url" : "https://gitlab.example.com/mmusterman/awesome_project.git",
         "id" : 239,
         "name" : "awesome_project",
         "namespace" : "mmusterman",
         "path_with_namespace" : "mmusterman/awesome_project",
         "ssh_url" : "git@gitlab.example.com:mmusterman/awesome_project.git",
         "url" : "git@gitlab.example.com:mmusterman/awesome_project.git",
         "visibility_level" : 0,
         "web_url" : "https://gitlab.example.com/mmusterman/awesome_project"
      },
      "target_branch" : "master",
      "target_project_id" : 239,
      "time_estimate" : 0,
      "title" : "Add note to README",
      "total_time_spent" : 0,
      "updated_at" : "2018-05-19 06:57:12 -0700",
      "updated_by_id" : null,
      "url" : "https://gitlab.example.com/mmusterman/awesome_project/merge_requests/7",
      "work_in_progress" : false
   },
   "object_kind" : "merge_request",
   "project" : {
      "avatar_url" : null,
      "ci_config_path" : null,
      "default_branch" : "master",
      "description" : "Trivial project for testing build machinery quickly",
      "git_http_url" : "https://gitlab.example.com/mmusterman/awesome_project.git",
      "git_ssh_url" : "git@gitlab.example.com:mmusterman/awesome_project.git",
      "homepage" : "https://gitlab.example.com/mmusterman/awesome_project",
      "http_url" : "https://gitlab.example.com/mmusterman/awesome_project.git",
      "id" : 239,
      "name" : "awesome_project",
      "namespace" : "mmusterman",
      "path_with_namespace" : "mmusterman/awesome_project",
      "ssh_url" : "git@gitlab.example.com:mmusterman/awesome_project.git",
      "url" : "git@gitlab.example.com:mmusterman/awesome_project.git",
      "visibility_level" : 0,
      "web_url" : "https://gitlab.example.com/mmusterman/awesome_project"
   },
   "repository" : {
      "description" : "Trivial project for testing build machinery quickly",
      "homepage" : "https://gitlab.example.com/mmusterman/awesome_project",
      "name" : "awesome_project",
      "url" : "git@gitlab.example.com:mmusterman/awesome_project.git"
   },
   "user" : {
      "avatar_url" : "http://www.gravatar.com/avatar/e64c7d89f26bd1972efa854d13d7dd61?s=40&d=identicon",
      "name" : "Max Mustermann",
      "username" : "mmusterman"
   }
}
"""  # noqa pylint: disable=line-too-long


def FakeRequestMR(content):
    request = FakeRequest(content=content)
    request.uri = b"/change_hook/gitlab"
    request.args = {b'codebase': [b'MyCodebase']}
    request.received_headers[_HEADER_EVENT] = b"Merge Request Hook"
    request.method = b"POST"
    return request


class TestChangeHookConfiguredWithGitChange(unittest.TestCase,
                                            TestReactorMixin):

    def setUp(self):
        self.setUpTestReactor()
        self.changeHook = change_hook.ChangeHookResource(
            dialects={'gitlab': True}, master=fakeMasterForHooks(self))

    def check_changes_tag_event(self, r, project='', codebase=None):
        self.assertEqual(len(self.changeHook.master.data.updates.changesAdded), 2)
        change = self.changeHook.master.data.updates.changesAdded[0]

        self.assertEqual(change["repository"], "git@localhost:diaspora.git")
        self.assertEqual(
            change["when_timestamp"],
            1323692851
        )
        self.assertEqual(change["branch"], "v1.0.0")

    def check_changes_mr_event(self, r, project='awesome_project', codebase=None,
                               timestamp=1526309644, source_repo=None):
        self.assertEqual(len(self.changeHook.master.data.updates.changesAdded), 1)
        change = self.changeHook.master.data.updates.changesAdded[0]

        self.assertEqual(change["repository"],
                         "https://gitlab.example.com/mmusterman/awesome_project.git")
        if source_repo is None:
            source_repo = "https://gitlab.example.com/mmusterman/awesome_project.git"
        self.assertEqual(change['properties']["source_repository"],
                         source_repo)
        self.assertEqual(change['properties']["target_repository"],
                         "https://gitlab.example.com/mmusterman/awesome_project.git")
        self.assertEqual(
            change["when_timestamp"],
            timestamp
        )
        self.assertEqual(change["branch"], "master")
        self.assertEqual(change['properties']["source_branch"], 'ms-viewport')
        self.assertEqual(change['properties']["target_branch"], 'master')
        self.assertEqual(change["category"], "merge_request")
        self.assertEqual(change.get("project"), project)

    def check_changes_push_event(self, r, project='diaspora', codebase=None):
        self.assertEqual(len(self.changeHook.master.data.updates.changesAdded), 2)
        change = self.changeHook.master.data.updates.changesAdded[0]

        self.assertEqual(change["repository"], "git@localhost:diaspora.git")
        self.assertEqual(
            change["when_timestamp"],
            1323692851
        )
        self.assertEqual(
            change["author"], "Jordi Mallach <jordi@softcatala.org>")
        self.assertEqual(
            change["revision"], 'b6568db1bc1dcd7f8b4d5a946b0b91f9dacd7327')
        self.assertEqual(
            change["comments"], "Update Catalan translation to e38cb41.")
        self.assertEqual(change["branch"], "master")
        self.assertEqual(change["revlink"],
            "http://localhost/diaspora/commits/b6568db1bc1dcd7f8b4d5a946b0b91f9dacd7327")

        change = self.changeHook.master.data.updates.changesAdded[1]
        self.assertEqual(change["repository"], "git@localhost:diaspora.git")
        self.assertEqual(
            change["when_timestamp"],
            1325626589
        )
        self.assertEqual(
            change["author"], "GitLab dev user <gitlabdev@dv6700.(none)>")
        self.assertEqual(change["src"], "git")
        self.assertEqual(
            change["revision"], 'da1560886d4f094c3e6c9ef40349f7d38b5d27d7')
        self.assertEqual(change["comments"], "fixed readme")
        self.assertEqual(change["branch"], "master")
        self.assertEqual(change["revlink"],
            "http://localhost/diaspora/commits/da1560886d4f094c3e6c9ef40349f7d38b5d27d7")

        # FIXME: should we convert project name to canonical case?
        # Or should change filter be case insensitive?
        self.assertEqual(change.get("project").lower(), project.lower())
        self.assertEqual(change.get("codebase"), codebase)

    # Test 'base' hook with attributes. We should get a json string representing
    # a Change object as a dictionary. All values show be set.
    @defer.inlineCallbacks
    def testGitWithChange(self):
        self.request = FakeRequest(content=gitJsonPayload)
        self.request.uri = b"/change_hook/gitlab"
        self.request.method = b"POST"
        self.request.received_headers[_HEADER_EVENT] = b"Push Hook"
        res = yield self.request.test_render(self.changeHook)
        self.check_changes_push_event(res)

    @defer.inlineCallbacks
    def testGitWithChange_WithProjectToo(self):
        self.request = FakeRequest(content=gitJsonPayload)
        self.request.uri = b"/change_hook/gitlab"
        self.request.args = {b'project': [b'Diaspora']}
        self.request.received_headers[_HEADER_EVENT] = b"Push Hook"
        self.request.method = b"POST"
        res = yield self.request.test_render(self.changeHook)
        self.check_changes_push_event(res, project="Diaspora")

    @defer.inlineCallbacks
    def testGitWithChange_WithCodebaseToo(self):
        self.request = FakeRequest(content=gitJsonPayload)
        self.request.uri = b"/change_hook/gitlab"
        self.request.args = {b'codebase': [b'MyCodebase']}
        self.request.received_headers[_HEADER_EVENT] = b"Push Hook"
        self.request.method = b"POST"
        res = yield self.request.test_render(self.changeHook)
        self.check_changes_push_event(res, codebase="MyCodebase")

    @defer.inlineCallbacks
    def testGitWithChange_WithPushTag(self):
        self.request = FakeRequest(content=gitJsonPayloadTag)
        self.request.uri = b"/change_hook/gitlab"
        self.request.args = {b'codebase': [b'MyCodebase']}
        self.request.received_headers[_HEADER_EVENT] = b"Push Hook"
        self.request.method = b"POST"
        res = yield self.request.test_render(self.changeHook)
        self.check_changes_tag_event(res, codebase="MyCodebase")

    @defer.inlineCallbacks
    def testGitWithNoJson(self):
        self.request = FakeRequest()
        self.request.uri = b"/change_hook/gitlab"
        self.request.method = b"POST"
        self.request.received_headers[_HEADER_EVENT] = b"Push Hook"
        yield self.request.test_render(self.changeHook)

        self.assertEqual(len(self.changeHook.master.data.updates.changesAdded), 0)
        self.assertIn(b"Error loading JSON:", self.request.written)
        self.request.setResponseCode.assert_called_with(400, mock.ANY)

    @defer.inlineCallbacks
    def test_event_property(self):
        self.request = FakeRequest(content=gitJsonPayload)
        self.request.received_headers[_HEADER_EVENT] = b"Push Hook"
        self.request.uri = b"/change_hook/gitlab"
        self.request.method = b"POST"
        yield self.request.test_render(self.changeHook)
        self.assertEqual(len(self.changeHook.master.data.updates.changesAdded), 2)
        change = self.changeHook.master.data.updates.changesAdded[0]
        self.assertEqual(change["properties"]["event"], "Push Hook")
        self.assertEqual(change["category"], "Push Hook")

    @defer.inlineCallbacks
    def testGitWithChange_WithMR_open(self):
        self.request = FakeRequestMR(content=gitJsonPayloadMR_open)
        res = yield self.request.test_render(self.changeHook)
        self.check_changes_mr_event(res, codebase="MyCodebase")
        change = self.changeHook.master.data.updates.changesAdded[0]
        self.assertEqual(change["category"], "merge_request")

    @defer.inlineCallbacks
    def testGitWithChange_WithMR_editdesc(self):
        self.request = FakeRequestMR(content=gitJsonPayloadMR_editdesc)
        yield self.request.test_render(self.changeHook)
        self.assertEqual(len(self.changeHook.master.data.updates.changesAdded), 0)

    @defer.inlineCallbacks
    def testGitWithChange_WithMR_addcommit(self):
        self.request = FakeRequestMR(content=gitJsonPayloadMR_addcommit)
        res = yield self.request.test_render(self.changeHook)
        self.check_changes_mr_event(res, codebase="MyCodebase", timestamp=1526395871)
        change = self.changeHook.master.data.updates.changesAdded[0]
        self.assertEqual(change["category"], "merge_request")

    @defer.inlineCallbacks
    def testGitWithChange_WithMR_close(self):
        self.request = FakeRequestMR(content=gitJsonPayloadMR_close)
        yield self.request.test_render(self.changeHook)
        self.assertEqual(len(self.changeHook.master.data.updates.changesAdded), 0)

    @defer.inlineCallbacks
    def testGitWithChange_WithMR_reopen(self):
        self.request = FakeRequestMR(content=gitJsonPayloadMR_reopen)
        res = yield self.request.test_render(self.changeHook)
        self.check_changes_mr_event(res, codebase="MyCodebase", timestamp=1526395871)
        change = self.changeHook.master.data.updates.changesAdded[0]
        self.assertEqual(change["category"], "merge_request")

    @defer.inlineCallbacks
    def testGitWithChange_WithMR_open_forked(self):
        self.request = FakeRequestMR(content=gitJsonPayloadMR_open_forked)
        res = yield self.request.test_render(self.changeHook)
        self.check_changes_mr_event(
                res, codebase="MyCodebase", timestamp=1526736926,
                source_repo="https://gitlab.example.com/build/awesome_project.git")
        change = self.changeHook.master.data.updates.changesAdded[0]
        self.assertEqual(change["category"], "merge_request")


class TestChangeHookConfiguredWithSecret(unittest.TestCase, TestReactorMixin):

    _SECRET = 'thesecret'

    def setUp(self):
        self.setUpTestReactor()
        self.master = fakeMasterForHooks(self)

        fakeStorageService = FakeSecretStorage()
        fakeStorageService.reconfigService(secretdict={"secret_key": self._SECRET})

        self.secretService = SecretManager()
        self.secretService.services = [fakeStorageService]
        self.master.addService(self.secretService)

        self.changeHook = change_hook.ChangeHookResource(
            dialects={'gitlab': {'secret': util.Secret("secret_key")}},
            master=self.master)

    @defer.inlineCallbacks
    def test_missing_secret(self):
        self.request = FakeRequest(content=gitJsonPayloadTag)
        self.request.uri = b"/change_hook/gitlab"
        self.request.args = {b'codebase': [b'MyCodebase']}
        self.request.method = b"POST"
        self.request.received_headers[_HEADER_EVENT] = b"Push Hook"
        yield self.request.test_render(self.changeHook)
        expected = b'Invalid secret'
        self.assertEqual(self.request.written, expected)
        self.assertEqual(len(self.changeHook.master.data.updates.changesAdded), 0)

    @defer.inlineCallbacks
    def test_valid_secret(self):
        self.request = FakeRequest(content=gitJsonPayload)
        self.request.received_headers[_HEADER_GITLAB_TOKEN] = self._SECRET
        self.request.received_headers[_HEADER_EVENT] = b"Push Hook"
        self.request.uri = b"/change_hook/gitlab"
        self.request.method = b"POST"
        yield self.request.test_render(self.changeHook)
        self.assertEqual(len(self.changeHook.master.data.updates.changesAdded), 2)



# This file is part of Buildbot.  Buildbot is free software: you can
# redistribute it and/or modify it under the terms of the GNU General Public
# License as published by the Free Software Foundation, version 2.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
# details.
#
# You should have received a copy of the GNU General Public License along with
# this program; if not, write to the Free Software Foundation, Inc., 51
# Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#
# Copyright Buildbot Team Members

import mock

from twisted.internet import defer
from twisted.trial import unittest

from buildbot.mq import simple
from buildbot.test.fake import fakemaster
from buildbot.test.util.misc import TestReactorMixin


class SimpleMQ(TestReactorMixin, unittest.TestCase):

    @defer.inlineCallbacks
    def setUp(self):
        self.setUpTestReactor()
        self.master = fakemaster.make_master(self)
        self.mq = simple.SimpleMQ()
        self.mq.setServiceParent(self.master)
        yield self.mq.startService()

    @defer.inlineCallbacks
    def tearDown(self):
        if self.mq.running:
            yield self.mq.stopService()

    @defer.inlineCallbacks
    def test_forward_data(self):
        callback = mock.Mock()
        yield self.mq.startConsuming(callback, ('a', 'b'))
        # _produce returns a deferred
        yield self.mq.produce(('a', 'b'), 'foo')
        # calling produce should eventually call the callback with decoding of
        # topic
        callback.assert_called_with(('a', 'b'), 'foo')

    @defer.inlineCallbacks
    def test_forward_data_wildcard(self):
        callback = mock.Mock()
        yield self.mq.startConsuming(callback, ('a', None))
        # _produce returns a deferred
        yield self.mq.produce(('a', 'b'), 'foo')
        # calling produce should eventually call the callback with decoding of
        # topic
        callback.assert_called_with(('a', 'b'), 'foo')

    @defer.inlineCallbacks
    def test_waits_for_called_callback(self):
        def callback(_, __):
            return defer.succeed(None)

        yield self.mq.startConsuming(callback, ('a', None))
        yield self.mq.produce(('a', 'b'), 'foo')

        d = self.mq.stopService()
        self.assertTrue(d.called)

    @defer.inlineCallbacks
    def test_waits_for_non_called_callback(self):
        d1 = defer.Deferred()

        def callback(_, __):
            return d1

        yield self.mq.startConsuming(callback, ('a', None))
        yield self.mq.produce(('a', 'b'), 'foo')

        d = self.mq.stopService()
        self.assertFalse(d.called)
        d1.callback(None)
        self.assertTrue(d.called)



# This file is part of Buildbot.  Buildbot is free software: you can
# redistribute it and/or modify it under the terms of the GNU General Public
# License as published by the Free Software Foundation, version 2.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
# details.
#
# You should have received a copy of the GNU General Public License along with
# this program; if not, write to the Free Software Foundation, Inc., 51
# Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#
# Copyright Buildbot Team Members

import re

from twisted.python import failure
from twisted.python import log

from buildbot import config
from buildbot.process import buildstep
from buildbot.process import logobserver
from buildbot.process import remotecommand
# for existing configurations that import WithProperties from here.  We like
# to move this class around just to keep our readers guessing.
from buildbot.process.properties import WithProperties
from buildbot.process.results import FAILURE
from buildbot.process.results import SUCCESS
from buildbot.process.results import WARNINGS
from buildbot.process.results import Results
from buildbot.process.results import worst_status
from buildbot.steps.worker import CompositeStepMixin
from buildbot.util import command_to_string
from buildbot.util import flatten
from buildbot.util import join_list

_hush_pyflakes = [WithProperties]
del _hush_pyflakes


class ShellCommand(buildstep.LoggingBuildStep):

    """I run a single shell command on the worker. I return FAILURE if
    the exit code of that command is non-zero, SUCCESS otherwise. To change
    this behavior, override my .evaluateCommand method, or customize
    decodeRC argument

    By default, a failure of this step will mark the whole build as FAILURE.
    To override this, give me an argument of flunkOnFailure=False .

    I create a single Log named 'log' which contains the output of the
    command. To create additional summary Logs, override my .createSummary
    method.

    The shell command I run (a list of argv strings) can be provided in
    several ways:
      - a class-level .command attribute
      - a command= parameter to my constructor (overrides .command)
      - set explicitly with my .setCommand() method (overrides both)

    @ivar command: a list of renderable objects (typically strings or
                   WithProperties instances). This will be used by start()
                   to create a RemoteShellCommand instance.

    @ivar logfiles: a dict mapping log NAMEs to workdir-relative FILENAMEs
                    of their corresponding logfiles. The contents of the file
                    named FILENAME will be put into a LogFile named NAME, ina
                    something approximating real-time. (note that logfiles=
                    is actually handled by our parent class LoggingBuildStep)

    @ivar lazylogfiles: Defaults to False. If True, logfiles will be tracked
                        `lazily', meaning they will only be added when and if
                        they are written to. Empty or nonexistent logfiles
                        will be omitted. (Also handled by class
                        LoggingBuildStep.)
    """

    name = "shell"
    renderables = [
        'command',
        'flunkOnFailure',
        'haltOnFailure',
        'remote_kwargs',
        'workerEnvironment'
    ]

    command = None  # set this to a command, or set in kwargs
    # logfiles={} # you can also set 'logfiles' to a dictionary, and it
    #               will be merged with any logfiles= argument passed in
    #               to __init__

    # override this on a specific ShellCommand if you want to let it fail
    # without dooming the entire build to a status of FAILURE
    flunkOnFailure = True

    def __init__(self, workdir=None,
                 command=None,
                 usePTY=None,
                 **kwargs):
        # most of our arguments get passed through to the RemoteShellCommand
        # that we create, but first strip out the ones that we pass to
        # BuildStep (like haltOnFailure and friends), and a couple that we
        # consume ourselves.

        if command:
            self.setCommand(command)

        if self.__class__ is ShellCommand and not command:
            # ShellCommand class is directly instantiated.
            # Explicitly check that command is set to prevent runtime error
            # later.
            config.error("ShellCommand's `command' argument is not specified")

        # pull out the ones that LoggingBuildStep wants, then upcall
        buildstep_kwargs = {}
        # workdir is here first positional argument, but it belongs to
        # BuildStep parent
        kwargs['workdir'] = workdir
        for k in list(kwargs):
            if k in self.__class__.parms:
                buildstep_kwargs[k] = kwargs[k]
                del kwargs[k]
        super().__init__(**buildstep_kwargs)

        # check validity of arguments being passed to RemoteShellCommand
        invalid_args = []
        valid_rsc_args = [
            'env',
            'want_stdout',
            'want_stderr',
            'timeout',
            'maxTime',
            'sigtermTime',
            'logfiles',
            'usePTY',
            'logEnviron',
            'collectStdout',
            'collectStderr',
            'interruptSignal',
            'initialStdin',
            'decodeRC',
            'stdioLogName',
        ]
        for arg in kwargs:
            if arg not in valid_rsc_args:
                invalid_args.append(arg)
        # Raise Configuration error in case invalid arguments are present
        if invalid_args:
            config.error("Invalid argument(s) passed to RemoteShellCommand: " +
                         ', '.join(invalid_args))

        # everything left over goes to the RemoteShellCommand
        kwargs['usePTY'] = usePTY
        self.remote_kwargs = kwargs
        self.remote_kwargs['workdir'] = workdir
        if self.__class__.__name__ not in ['TreeSize', 'SetPropertyFromCommand', 'Configure',
                                           'WarningCountingShellCommand', 'Compile', 'Test',
                                           'MTR', 'DebPbuilder', 'DebCowbuilder', 'UbuPbuilder']:
            self.warn_deprecated_if_oldstyle_subclass('ShellCommand')

    def setBuild(self, build):
        super().setBuild(build)
        # Set this here, so it gets rendered when we start the step
        self.workerEnvironment = self.build.workerEnvironment

    def setCommand(self, command):
        self.command = command

    def _describe(self, done=False):
        return None

    def describe(self, done=False):
        if self.stopped and not self.rendered:
            return "stopped early"
        assert(self.rendered)
        desc = self._describe(done)
        if not desc:
            return None
        if self.descriptionSuffix:
            desc = desc + ' ' + join_list(self.descriptionSuffix)
        return desc

    def getCurrentSummary(self):
        cmdsummary = self._getLegacySummary(False)
        if cmdsummary:
            return {'step': cmdsummary}
        return super().getCurrentSummary()

    def getResultSummary(self):
        cmdsummary = self._getLegacySummary(True)

        if cmdsummary:
            if self.results != SUCCESS:
                cmdsummary += ' ({})'.format(Results[self.results])
            return {'step': cmdsummary}

        return super().getResultSummary()

    def _getLegacySummary(self, done):
        # defer to the describe method, if set
        description = self.describe(done)
        if description:
            return join_list(description)

        # defer to descriptions, if they're set
        if (not done and self.description) or (done and self.descriptionDone):
            return None

        try:
            # if self.cmd is set, then use the RemoteCommand's info
            if self.cmd:
                command = self.cmd.remote_command
            # otherwise, if we were configured with a command, use that
            elif self.command:
                command = self.command
            else:
                return None

            rv = command_to_string(command)

            # add the descriptionSuffix, if one was given
            if self.descriptionSuffix:
                rv = rv + ' ' + join_list(self.descriptionSuffix)

            return rv

        except Exception:
            log.err(failure.Failure(), "Error describing step")
            return None

    def setupEnvironment(self, cmd):
        # merge in anything from workerEnvironment (which comes from the builder
        # config) Environment variables passed in by a BuildStep override those
        # passed in at the Builder level, so if we have any from the builder,
        # apply those and then update with the args from the buildstep
        # (cmd.args)
        workerEnv = self.workerEnvironment
        if workerEnv:
            if cmd.args['env'] is None:
                cmd.args['env'] = {}
            fullWorkerEnv = workerEnv.copy()
            fullWorkerEnv.update(cmd.args['env'])
            cmd.args['env'] = fullWorkerEnv
            # note that each RemoteShellCommand gets its own copy of the
            # dictionary, so we shouldn't be affecting anyone but ourselves.

    def buildCommandKwargs(self, warnings):
        kwargs = super().buildCommandKwargs()
        kwargs.update(self.remote_kwargs)
        kwargs['workdir'] = self.workdir

        kwargs['command'] = flatten(self.command, (list, tuple))

        # check for the usePTY flag
        if 'usePTY' in kwargs and kwargs['usePTY'] is not None:
            if self.workerVersionIsOlderThan("shell", "2.7"):
                warnings.append(
                    "NOTE: worker does not allow master to override usePTY\n")
                del kwargs['usePTY']

        # check for the interruptSignal flag
        if "interruptSignal" in kwargs and self.workerVersionIsOlderThan("shell", "2.15"):
            warnings.append(
                "NOTE: worker does not allow master to specify interruptSignal\n")
            del kwargs['interruptSignal']

        return kwargs

    def start(self):
        # this block is specific to ShellCommands. subclasses that don't need
        # to set up an argv array, an environment, or extra logfiles= (like
        # the Source subclasses) can just skip straight to startCommand()

        warnings = []

        # create the actual RemoteShellCommand instance now
        kwargs = self.buildCommandKwargs(warnings)
        cmd = remotecommand.RemoteShellCommand(**kwargs)
        self.setupEnvironment(cmd)

        self.startCommand(cmd, warnings)


class TreeSize(ShellCommand):
    name = "treesize"
    command = ["du", "-s", "-k", "."]
    description = "measuring tree size"
    kib = None

    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self.observer = logobserver.BufferLogObserver(wantStdout=True,
                                                      wantStderr=True)
        self.addLogObserver('stdio', self.observer)

    def commandComplete(self, cmd):
        out = self.observer.getStdout()
        m = re.search(r'^(\d+)', out)
        if m:
            self.kib = int(m.group(1))
            self.setProperty("tree-size-KiB", self.kib, "treesize")

    def evaluateCommand(self, cmd):
        if cmd.didFail():
            return FAILURE
        if self.kib is None:
            return WARNINGS  # not sure how 'du' could fail, but whatever
        return SUCCESS

    def _describe(self, done=False):
        if self.kib is not None:
            return ["treesize", "%d KiB" % self.kib]
        return ["treesize", "unknown"]


class SetPropertyFromCommand(ShellCommand):
    name = "setproperty"
    renderables = ['property']

    def __init__(self, property=None, extract_fn=None, strip=True,
                 includeStdout=True, includeStderr=False, **kwargs):
        self.property = property
        self.extract_fn = extract_fn
        self.strip = strip
        self.includeStdout = includeStdout
        self.includeStderr = includeStderr

        if not ((property is not None) ^ (extract_fn is not None)):
            config.error(
                "Exactly one of property and extract_fn must be set")

        super().__init__(**kwargs)

        if self.extract_fn:
            self.includeStderr = True

        self.observer = logobserver.BufferLogObserver(
            wantStdout=self.includeStdout,
            wantStderr=self.includeStderr)
        self.addLogObserver('stdio', self.observer)

        self.property_changes = {}

        self.warn_deprecated_if_oldstyle_subclass('SetPropertyFromCommand')

    def commandComplete(self, cmd):
        if self.property:
            if cmd.didFail():
                return
            result = self.observer.getStdout()
            if self.strip:
                result = result.strip()
            propname = self.property
            self.setProperty(propname, result, "SetPropertyFromCommand Step")
            self.property_changes[propname] = result
        else:
            new_props = self.extract_fn(cmd.rc,
                                        self.observer.getStdout(),
                                        self.observer.getStderr())
            for k, v in new_props.items():
                self.setProperty(k, v, "SetPropertyFromCommand Step")
            self.property_changes = new_props

    def createSummary(self, log):
        if self.property_changes:
            props_set = ["{}: {}".format(k, repr(v))
                         for k, v in sorted(self.property_changes.items())]
            self.addCompleteLog('property changes', "\n".join(props_set))

    def describe(self, done=False):
        if len(self.property_changes) > 1:
            return ["%d properties set" % len(self.property_changes)]
        elif len(self.property_changes) == 1:
            return ["property '{}' set".format(list(self.property_changes)[0])]
        # else:
        # let ShellCommand describe
        return super().describe(done)


class Configure(ShellCommand):

    name = "configure"
    haltOnFailure = 1
    flunkOnFailure = 1
    description = ["configuring"]
    descriptionDone = ["configure"]
    command = ["./configure"]

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.warn_deprecated_if_oldstyle_subclass('Configure')


class WarningCountingShellCommand(ShellCommand, CompositeStepMixin):
    renderables = [
                    'suppressionFile',
                    'suppressionList',
                    'warningPattern',
                    'directoryEnterPattern',
                    'directoryLeavePattern',
                    'maxWarnCount',
    ]

    warnCount = 0
    warningPattern = '(?i).*warning[: ].*'
    # The defaults work for GNU Make.
    directoryEnterPattern = ("make.*: Entering directory "
                             "[\u2019\"`'](.*)[\u2019'`\"]")
    directoryLeavePattern = "make.*: Leaving directory"
    suppressionFile = None

    commentEmptyLineRe = re.compile(r"^\s*(#.*)?$")
    suppressionLineRe = re.compile(
        r"^\s*(.+?)\s*:\s*(.+?)\s*(?:[:]\s*([0-9]+)(?:-([0-9]+))?\s*)?$")

    def __init__(self,
                 warningPattern=None, warningExtractor=None, maxWarnCount=None,
                 directoryEnterPattern=None, directoryLeavePattern=None,
                 suppressionFile=None, suppressionList=None, **kwargs):
        # See if we've been given a regular expression to use to match
        # warnings. If not, use a default that assumes any line with "warning"
        # present is a warning. This may lead to false positives in some cases.
        if warningPattern:
            self.warningPattern = warningPattern
        if directoryEnterPattern:
            self.directoryEnterPattern = directoryEnterPattern
        if directoryLeavePattern:
            self.directoryLeavePattern = directoryLeavePattern
        if suppressionFile:
            self.suppressionFile = suppressionFile
        # self.suppressions is already taken, so use something else
        self.suppressionList = suppressionList
        if warningExtractor:
            self.warningExtractor = warningExtractor
        else:
            self.warningExtractor = WarningCountingShellCommand.warnExtractWholeLine
        self.maxWarnCount = maxWarnCount

        # And upcall to let the base class do its work
        super().__init__(**kwargs)

        if self.__class__ is WarningCountingShellCommand and \
                not kwargs.get('command'):
            # WarningCountingShellCommand class is directly instantiated.
            # Explicitly check that command is set to prevent runtime error
            # later.
            config.error("WarningCountingShellCommand's `command' argument "
                         "is not specified")

        self.suppressions = []
        self.directoryStack = []

        self.warnCount = 0
        self.loggedWarnings = []

        self.addLogObserver(
            'stdio',
            logobserver.LineConsumerLogObserver(self.warningLogConsumer))

        if self.__class__.__name__ not in ['Compile', 'Test', 'DebPbuilder', 'DebCowbuilder',
                                           'UbuPbuilder']:
            self.warn_deprecated_if_oldstyle_subclass('WarningCountingShellCommand')

    def addSuppression(self, suppressionList):
        """
        This method can be used to add patters of warnings that should
        not be counted.

        It takes a single argument, a list of patterns.

        Each pattern is a 4-tuple (FILE-RE, WARN-RE, START, END).

        FILE-RE is a regular expression (string or compiled regexp), or None.
        If None, the pattern matches all files, else only files matching the
        regexp. If directoryEnterPattern is specified in the class constructor,
        matching is against the full path name, eg. src/main.c.

        WARN-RE is similarly a regular expression matched against the
        text of the warning, or None to match all warnings.

        START and END form an inclusive line number range to match against. If
        START is None, there is no lower bound, similarly if END is none there
        is no upper bound."""

        for fileRe, warnRe, start, end in suppressionList:
            if fileRe is not None and isinstance(fileRe, str):
                fileRe = re.compile(fileRe)
            if warnRe is not None and isinstance(warnRe, str):
                warnRe = re.compile(warnRe)
            self.suppressions.append((fileRe, warnRe, start, end))

    def warnExtractWholeLine(self, line, match):
        """
        Extract warning text as the whole line.
        No file names or line numbers."""
        return (None, None, line)

    def warnExtractFromRegexpGroups(self, line, match):
        """
        Extract file name, line number, and warning text as groups (1,2,3)
        of warningPattern match."""
        file = match.group(1)
        lineNo = match.group(2)
        if lineNo is not None:
            lineNo = int(lineNo)
        text = match.group(3)
        return (file, lineNo, text)

    def warningLogConsumer(self):
        # Now compile a regular expression from whichever warning pattern we're
        # using
        wre = self.warningPattern
        if isinstance(wre, str):
            wre = re.compile(wre)

        directoryEnterRe = self.directoryEnterPattern
        if (directoryEnterRe is not None and
                isinstance(directoryEnterRe, str)):
            directoryEnterRe = re.compile(directoryEnterRe)

        directoryLeaveRe = self.directoryLeavePattern
        if (directoryLeaveRe is not None and
                isinstance(directoryLeaveRe, str)):
            directoryLeaveRe = re.compile(directoryLeaveRe)

        # Check if each line in the output from this command matched our
        # warnings regular expressions. If did, bump the warnings count and
        # add the line to the collection of lines with warnings
        self.loggedWarnings = []
        while True:
            stream, line = yield
            if directoryEnterRe:
                match = directoryEnterRe.search(line)
                if match:
                    self.directoryStack.append(match.group(1))
                    continue
            if (directoryLeaveRe and
                self.directoryStack and
                    directoryLeaveRe.search(line)):
                self.directoryStack.pop()
                continue

            match = wre.match(line)
            if match:
                self.maybeAddWarning(self.loggedWarnings, line, match)

    def maybeAddWarning(self, warnings, line, match):
        if self.suppressions:
            (file, lineNo, text) = self.warningExtractor(self, line, match)
            lineNo = lineNo and int(lineNo)

            if file is not None and file != "" and self.directoryStack:
                currentDirectory = '/'.join(self.directoryStack)
                if currentDirectory is not None and currentDirectory != "":
                    file = "{}/{}".format(currentDirectory, file)

            # Skip adding the warning if any suppression matches.
            for fileRe, warnRe, start, end in self.suppressions:
                if not (file is None or fileRe is None or fileRe.match(file)):
                    continue
                if not (warnRe is None or warnRe.search(text)):
                    continue
                if ((start is not None and end is not None) and
                   not (lineNo is not None and start <= lineNo <= end)):
                    continue
                return

        warnings.append(line)
        self.warnCount += 1

    def start(self):
        if self.suppressionList is not None:
            self.addSuppression(self.suppressionList)
        if self.suppressionFile is None:
            return super().start()
        d = self.getFileContentFromWorker(
            self.suppressionFile, abandonOnFailure=True)
        d.addCallback(self.uploadDone)
        d.addErrback(self.failed)
        return None

    def uploadDone(self, data):
        lines = data.split("\n")

        list = []
        for line in lines:
            if self.commentEmptyLineRe.match(line):
                continue
            match = self.suppressionLineRe.match(line)
            if (match):
                file, test, start, end = match.groups()
                if (end is not None):
                    end = int(end)
                if (start is not None):
                    start = int(start)
                    if end is None:
                        end = start
                list.append((file, test, start, end))

        self.addSuppression(list)
        return super().start()

    def createSummary(self, log):
        """
        Match log lines against warningPattern.

        Warnings are collected into another log for this step, and the
        build-wide 'warnings-count' is updated."""

        # If there were any warnings, make the log if lines with warnings
        # available
        if self.warnCount:
            self.addCompleteLog("warnings (%d)" % self.warnCount,
                                "\n".join(self.loggedWarnings) + "\n")

        warnings_stat = self.getStatistic('warnings', 0)
        self.setStatistic('warnings', warnings_stat + self.warnCount)

        old_count = self.getProperty("warnings-count", 0)
        self.setProperty(
            "warnings-count", old_count + self.warnCount, "WarningCountingShellCommand")

    def evaluateCommand(self, cmd):
        result = cmd.results()
        if (self.maxWarnCount is not None and self.warnCount > self.maxWarnCount):
            result = worst_status(result, FAILURE)
        elif self.warnCount:
            result = worst_status(result, WARNINGS)
        return result


class Compile(WarningCountingShellCommand):

    name = "compile"
    haltOnFailure = 1
    flunkOnFailure = 1
    description = ["compiling"]
    descriptionDone = ["compile"]
    command = ["make", "all"]

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.warn_deprecated_if_oldstyle_subclass('Compile')


class Test(WarningCountingShellCommand):

    name = "test"
    warnOnFailure = 1
    description = ["testing"]
    descriptionDone = ["test"]
    command = ["make", "test"]

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        if self.__class__.__name__ not in ['MTR']:
            self.warn_deprecated_if_oldstyle_subclass('Test')

    def setTestResults(self, total=0, failed=0, passed=0, warnings=0):
        """
        Called by subclasses to set the relevant statistics; this actually
        adds to any statistics already present
        """
        total += self.getStatistic('tests-total', 0)
        self.setStatistic('tests-total', total)
        failed += self.getStatistic('tests-failed', 0)
        self.setStatistic('tests-failed', failed)
        warnings += self.getStatistic('tests-warnings', 0)
        self.setStatistic('tests-warnings', warnings)
        passed += self.getStatistic('tests-passed', 0)
        self.setStatistic('tests-passed', passed)

    def describe(self, done=False):
        description = super().describe(done)
        if done:
            if not description:
                description = []
            description = description[:]  # make a private copy
            if self.hasStatistic('tests-total'):
                total = self.getStatistic("tests-total", 0)
                failed = self.getStatistic("tests-failed", 0)
                passed = self.getStatistic("tests-passed", 0)
                warnings = self.getStatistic("tests-warnings", 0)
                if not total:
                    total = failed + passed + warnings

                if total:
                    description.append('%d tests' % total)
                if passed:
                    description.append('%d passed' % passed)
                if warnings:
                    description.append('%d warnings' % warnings)
                if failed:
                    description.append('%d failed' % failed)
        return description



# This file is part of Buildbot.  Buildbot is free software: you can
# redistribute it and/or modify it under the terms of the GNU General Public
# License as published by the Free Software Foundation, version 2.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
# details.
#
# You should have received a copy of the GNU General Public License along with
# this program; if not, write to the Free Software Foundation, Inc., 51
# Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#
# Copyright Buildbot Team Members

import re
from datetime import datetime

from twisted.internet import defer
from twisted.trial import unittest
from twisted.web import client
from twisted.web.error import Error

from buildbot.changes.bitbucket import BitbucketPullrequestPoller
from buildbot.test.util import changesource
from buildbot.test.util.misc import TestReactorMixin


class SourceRest():
    """https://bitbucket.org/!api/2.0/repositories/{owner}/{slug}"""
    template = """\
{

    "hash": "%(hash)s",
    "links": {
        "html": {
            "href": "https://bitbucket.org/%(owner)s/%(slug)s/commits/%(short_hash)s"
        }
    },
    "repository": {
        "links": {
            "self": {
                "href": "https://bitbucket.org/!api/2.0/repositories/%(owner)s/%(slug)s"
            }
        }
    },
    "date": "%(date)s"

}
"""
    repo_template = """\
{
    "links": {
        "html": {
            "href": "https://bitbucket.org/%(owner)s/%(slug)s"
        }
    }
}
"""

    def __init__(self, owner, slug, hash, date):
        self.owner = owner
        self.slug = slug
        self.hash = hash
        self.date = date

    def request(self):
        return self.template % {
            "owner": self.owner,
            "slug": self.slug,
            "hash": self.hash,
            "short_hash": self.hash[0:12],
            "date": self.date,
        }

    def repo_request(self):
        return self.repo_template % {
            "owner": self.owner,
            "slug": self.slug,
        }


class PullRequestRest():
    """https://bitbucket.org/!api/2.0/repositories/{owner}/{slug}/pullrequests/{pull_request_id}"""
    template = """\
{

    "description": "%(description)s",
    "title": "%(title)s",
    "source": {
        "commit": {
            "hash": "%(hash)s",
            "links": {
                "self": {
                    "href": "https://bitbucket.org/!api/2.0/repositories/%(owner)s/%(slug)s/commit/%(hash)s"
                }
            }
        }
    },
    "state": "OPEN",
    "author": {
        "display_name": "%(display_name)s"
    },
    "created_on": "%(created_on)s",
    "participants": [
    ],
    "updated_on": "%(updated_on)s",
    "merge_commit": null,
    "id": %(id)d

}
"""  # noqa pylint: disable=line-too-long

    def __init__(self, nr, title, description, display_name, source, created_on, updated_on=None):
        self.nr = nr
        self.title = title
        self.description = description
        self.display_name = display_name
        self.source = source
        self.created_on = created_on
        if updated_on:
            self.updated_on = updated_on
        else:
            self.updated_on = self.created_on

    def request(self):
        return self.template % {
            "description": self.description,
            "title": self.title,
            "hash": self.source.hash,
            "short_hash": self.source.hash[0:12],
            "owner": self.source.owner,
            "slug": self.source.slug,
            "display_name": self.display_name,
            "created_on": self.created_on,
            "updated_on": self.updated_on,
            "id": self.nr,
        }


class PullRequestListRest():
    """https://bitbucket.org/api/2.0/repositories/{owner}/{slug}/pullrequests"""
    template = """\
        {
            "description": "%(description)s",
            "links": {
                "self": {
                    "href": "https://bitbucket.org/!api/2.0/repositories/%(owner)s/%(slug)s/pullrequests/%(id)d"
                },
                "html": {
                    "href": "https://bitbucket.org/%(owner)s/%(slug)s/pull-request/%(id)d"
                }
            },
            "author": {
                "display_name": "%(display_name)s"
            },
            "title": "%(title)s",
            "source": {
                "commit": {
                    "hash": "%(short_hash)s",
                    "links": {
                        "self": {
                            "href": "https://bitbucket.org/!api/2.0/repositories/%(src_owner)s/%(src_slug)s/commit/%(short_hash)s"
                        }
                    }
                },
                "repository": {
                    "links": {
                        "self": {
                            "href": "https://bitbucket.org/!api/2.0/repositories/%(src_owner)s/%(src_slug)s"
                        }
                    }
                },
                "branch": {
                    "name": "default"
                }
            },
            "state": "OPEN",
            "created_on": "%(created_on)s",
            "updated_on": "%(updated_on)s",
            "merge_commit": null,
            "id": %(id)s
        }
"""  # noqa pylint: disable=line-too-long

    def __init__(self, owner, slug, prs):
        self.owner = owner
        self.slug = slug
        self.prs = prs

        self.pr_by_id = {}
        self.src_by_url = {}
        for pr in prs:
            self.pr_by_id[pr.nr] = pr
            self.src_by_url["{}/{}".format(pr.source.owner, pr.source.slug)] = pr.source

    def request(self):

        s = ""
        for pr in self.prs:
            s += self.template % {
                "description": pr.description,
                "owner": self.owner,
                "slug": self.slug,
                "display_name": pr.display_name,
                "title": pr.title,
                "hash": pr.source.hash,
                "short_hash": pr.source.hash[0:12],
                "src_owner": pr.source.owner,
                "src_slug": pr.source.slug,
                "created_on": pr.created_on,
                "updated_on": pr.updated_on,
                "id": pr.nr,
            }
        return """\
{

    "pagelen": 10,
    "values": [%s
    ],
    "page": 1

}
""" % s

    def getPage(self, url, timeout=None):

        list_url_re = re.compile(
            r"https://bitbucket.org/api/2.0/repositories/{}/{}/pullrequests".format(self.owner,
                                                                                    self.slug))
        pr_url_re = re.compile(
            r"https://bitbucket.org/!api/2.0/repositories/{}/{}/pullrequests/(?P<id>\d+)".format(
                    self.owner, self.slug))
        source_commit_url_re = re.compile(
            r"https://bitbucket.org/!api/2.0/repositories/(?P<src_owner>.*)/(?P<src_slug>.*)/commit/(?P<hash>\d+)")  # noqa pylint: disable=line-too-long
        source_url_re = re.compile(
            r"https://bitbucket.org/!api/2.0/repositories/(?P<src_owner>.*)/(?P<src_slug>.*)")

        if list_url_re.match(url):
            return defer.succeed(self.request())

        m = pr_url_re.match(url)
        if m:
            return self.pr_by_id[int(m.group("id"))].request()

        m = source_commit_url_re.match(url)
        if m:
            return self.src_by_url["{}/{}".format(m.group("src_owner"),
                                                  m.group("src_slug"))].request()

        m = source_url_re.match(url)
        if m:
            return self.src_by_url["{}/{}".format(m.group("src_owner"),
                                                  m.group("src_slug"))].repo_request()

        raise Error(code=404)


class TestBitbucketPullrequestPoller(changesource.ChangeSourceMixin,
                                     TestReactorMixin,
                                     unittest.TestCase):

    def setUp(self):
        self.setUpTestReactor()

        # create pull requests
        self.date = "2013-10-15T20:38:20.001797+00:00"
        self.date_epoch = datetime.strptime(self.date.split('.')[0],
                                            '%Y-%m-%dT%H:%M:%S')
        src = SourceRest(
            owner="contributor",
            slug="slug",
            hash="1111111111111111111111111111111111111111",
            date=self.date,
        )
        pr = PullRequestRest(
            nr=1,
            title="title",
            description="description",
            display_name="contributor",
            source=src,
            created_on=self.date,
        )
        self.pr_list = PullRequestListRest(
            owner="owner",
            slug="slug",
            prs=[pr],
        )
        # update
        src = SourceRest(
            owner="contributor",
            slug="slug",
            hash="2222222222222222222222222222222222222222",
            date=self.date,
        )
        pr = PullRequestRest(
            nr=1,
            title="title",
            description="description",
            display_name="contributor",
            source=src,
            created_on=self.date,
        )
        self.pr_list2 = PullRequestListRest(
            owner="owner",
            slug="slug",
            prs=[pr],
        )
        return self.setUpChangeSource()

    def tearDown(self):
        return self.tearDownChangeSource()

    def _fakeGetPage(self, result):
        # Install a fake getPage that puts the requested URL in self.getPage_got_url
        # and return result
        self.getPage_got_url = None

        def fake(url, timeout=None):
            self.getPage_got_url = url
            return defer.succeed(result)
        self.patch(client, "getPage", fake)

    def _fakeGetPage404(self):

        def fail(url, timeout=None):
            raise Error(code=404)
        self.patch(client, "getPage", fail)

    def attachDefaultChangeSource(self):
        return self.attachChangeSource(BitbucketPullrequestPoller(
            owner='owner',
            slug='slug'))

    # tests
    @defer.inlineCallbacks
    def test_describe(self):
        yield self.attachDefaultChangeSource()
        assert re.search(r'owner/slug', self.changesource.describe())

    @defer.inlineCallbacks
    def test_poll_unknown_repo(self):
        yield self.attachDefaultChangeSource()
        # Polling a non-existent repository should result in a 404
        self._fakeGetPage404()
        try:
            yield self.changesource.poll()
            self.fail(
                'Polling a non-existent repository should result in a 404.')
        except Exception as e:
            self.assertEqual(str(e), '404 Not Found')

    @defer.inlineCallbacks
    def test_poll_no_pull_requests(self):
        yield self.attachDefaultChangeSource()
        rest = PullRequestListRest(owner="owner", slug="slug", prs=[])
        self._fakeGetPage(rest.request())
        yield self.changesource.poll()

        self.assertEqual(len(self.master.data.updates.changesAdded), 0)

    @defer.inlineCallbacks
    def test_poll_new_pull_requests(self):
        yield self.attachDefaultChangeSource()
        # patch client.getPage()
        self.patch(client, "getPage", self.pr_list.getPage)

        yield self.changesource.poll()

        self.assertEqual(self.master.data.updates.changesAdded, [{
            'author': 'contributor',
            'committer': None,
            'branch': 'default',
            'category': None,
            'codebase': None,
            'comments': 'pull-request #1: title\nhttps://bitbucket.org/owner/slug/pull-request/1',
            'files': None,
            'project': '',
            'properties': {},
            'repository': 'https://bitbucket.org/contributor/slug',
            'revision': '1111111111111111111111111111111111111111',
            'revlink': 'https://bitbucket.org/contributor/slug/commits/111111111111',
            'src': 'bitbucket',
            'when_timestamp': 1381869500,
        }])

    @defer.inlineCallbacks
    def test_poll_no_updated_pull_request(self):
        yield self.attachDefaultChangeSource()

        # patch client.getPage()
        self.patch(client, "getPage", self.pr_list.getPage)

        yield self.changesource.poll()

        self.assertEqual(self.master.data.updates.changesAdded, [{
            'author': 'contributor',
            'committer': None,
            'branch': 'default',
            'category': None,
            'codebase': None,
            'comments': 'pull-request #1: title\nhttps://bitbucket.org/owner/slug/pull-request/1',
            'files': None,
            'project': '',
            'properties': {},
            'repository': 'https://bitbucket.org/contributor/slug',
            'revision': '1111111111111111111111111111111111111111',
            'revlink': 'https://bitbucket.org/contributor/slug/commits/111111111111',
            'src': 'bitbucket',
            'when_timestamp': 1381869500,
        }])

        # repoll
        yield self.changesource.poll()
        self.assertEqual(len(self.master.data.updates.changesAdded), 1)

    @defer.inlineCallbacks
    def test_poll_updated_pull_request(self):
        yield self.attachDefaultChangeSource()
        # patch client.getPage()
        self.patch(client, "getPage", self.pr_list.getPage)

        yield self.changesource.poll()

        self.assertEqual(self.master.data.updates.changesAdded, [{
            'author': 'contributor',
            'committer': None,
            'branch': 'default',
            'category': None,
            'codebase': None,
            'comments': 'pull-request #1: title\nhttps://bitbucket.org/owner/slug/pull-request/1',
            'files': None,
            'project': '',
            'properties': {},
            'repository': 'https://bitbucket.org/contributor/slug',

            'revision': '1111111111111111111111111111111111111111',
            'revlink': 'https://bitbucket.org/contributor/slug/commits/111111111111',
            'src': 'bitbucket',
            'when_timestamp': 1381869500,
        }])
        self.patch(client, "getPage", self.pr_list2.getPage)
        yield self.changesource.poll()

        self.assertEqual(self.master.data.updates.changesAdded, [
            {
                'author': 'contributor',
                'committer': None,
                'branch': 'default',
                'category': None,
                'codebase': None,
                'comments':
                    'pull-request #1: title\nhttps://bitbucket.org/owner/slug/pull-request/1',
                'files': None,
                'project': '',
                'properties': {},
                'repository': 'https://bitbucket.org/contributor/slug',
                'revision': '1111111111111111111111111111111111111111',
                'revlink': 'https://bitbucket.org/contributor/slug/commits/111111111111',
                'src': 'bitbucket',
                'when_timestamp': 1381869500,
            },
            {
                'author': 'contributor',
                'committer': None,
                'branch': 'default',
                'category': None,
                'codebase': None,
                'comments':
                    'pull-request #1: title\nhttps://bitbucket.org/owner/slug/pull-request/1',
                'files': None,
                'project': '',
                'properties': {},
                'repository': 'https://bitbucket.org/contributor/slug',
                'revision': '2222222222222222222222222222222222222222',
                'revlink': 'https://bitbucket.org/contributor/slug/commits/222222222222',
                'src': 'bitbucket',
                'when_timestamp': 1381869500,
            }
        ])

    @defer.inlineCallbacks
    def test_poll_pull_request_filter_False(self):
        yield self.attachChangeSource(BitbucketPullrequestPoller(
            owner='owner',
            slug='slug',
            pullrequest_filter=lambda x: False
        ))

        # patch client.getPage()
        self.patch(client, "getPage", self.pr_list.getPage)

        yield self.changesource.poll()

        self.assertEqual(len(self.master.data.updates.changesAdded), 0)

    @defer.inlineCallbacks
    def test_poll_pull_request_filter_True(self):
        yield self.attachChangeSource(BitbucketPullrequestPoller(
            owner='owner',
            slug='slug',
            pullrequest_filter=lambda x: True
        ))

        # patch client.getPage()
        self.patch(client, "getPage", self.pr_list.getPage)

        yield self.changesource.poll()

        self.assertEqual(self.master.data.updates.changesAdded, [{
            'author': 'contributor',
            'committer': None,
            'branch': 'default',
            'category': None,
            'codebase': None,
            'comments': 'pull-request #1: title\nhttps://bitbucket.org/owner/slug/pull-request/1',
            'files': None,
            'project': '',
            'properties': {},
            'repository': 'https://bitbucket.org/contributor/slug',
            'revision': '1111111111111111111111111111111111111111',
            'revlink': 'https://bitbucket.org/contributor/slug/commits/111111111111',
            'src': 'bitbucket',
            'when_timestamp': 1381869500,
        }])

    @defer.inlineCallbacks
    def test_poll_pull_request_not_useTimestamps(self):
        yield self.attachChangeSource(BitbucketPullrequestPoller(
            owner='owner',
            slug='slug',
            useTimestamps=False,
        ))

        self.patch(client, "getPage", self.pr_list.getPage)
        self.reactor.advance(1396825656)

        yield self.changesource.poll()
        self.assertEqual(self.master.data.updates.changesAdded, [{
            'author': 'contributor',
            'committer': None,
            'branch': 'default',
            'category': None,
            'codebase': None,
            'comments': 'pull-request #1: title\nhttps://bitbucket.org/owner/slug/pull-request/1',
            'files': None,
            'project': '',
            'properties': {},
            'repository': 'https://bitbucket.org/contributor/slug',
            'revision': '1111111111111111111111111111111111111111',
            'revlink': 'https://bitbucket.org/contributor/slug/commits/111111111111',
            'src': 'bitbucket',
            'when_timestamp': 1396825656,
        }])



# This file is part of Buildbot.  Buildbot is free software: you can
# redistribute it and/or modify it under the terms of the GNU General Public
# License as published by the Free Software Foundation, version 2.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
# details.
#
# You should have received a copy of the GNU General Public License along with
# this program; if not, write to the Free Software Foundation, Inc., 51
# Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#
# Copyright Buildbot Team Members

import os
import unittest
from builtins import int

from twisted.python import util


def onlyOnce(fn):
    'Set up FN to only run once within an interpreter instance'
    def wrap(*args, **kwargs):
        if hasattr(fn, 'called'):
            return None
        fn.called = 1
        return fn(*args, **kwargs)
    util.mergeFunctionMetadata(fn, wrap)
    return wrap

# NOTE: all of these patches test for applicability *before* importing the
# patch module.  This will help cut down on unnecessary imports where the
# patches are not needed, and also avoid problems with patches importing
# private things in external libraries that no longer exist.


@onlyOnce
def patch_testcase_timeout():
    # any test that should take more than 5 second should be annotated so.
    unittest.TestCase.timeout = 5

    # but we know that the DB tests are very slow, so we increase a bit that value for
    # real database tests
    if os.environ.get("BUILDBOT_TEST_DB_URL", None) is not None:
        unittest.TestCase.timeout = 120


@onlyOnce
def patch_servicechecks():
    from buildbot.monkeypatches import servicechecks
    servicechecks.patch()


@onlyOnce
def patch_mysqlclient_warnings():
    try:
        from _mysql_exceptions import Warning
        # MySQLdb.compat is only present in mysqlclient
        import MySQLdb.compat  # noqa pylint: disable=unused-import,import-outside-toplevel
    except ImportError:
        return
    # workaround for https://twistedmatrix.com/trac/ticket/9005
    # mysqlclient is easier to patch than twisted
    # we swap _mysql_exceptions.Warning arguments so that the code is in second place

    def patched_init(self, *args):
        if isinstance(args[0], int):
            super().__init__("{} {}".format(args[0], args[1]))
        else:
            super().__init__(*args)
    Warning.__init__ = patched_init


@onlyOnce
def patch_decorators():
    from buildbot.monkeypatches import decorators
    decorators.patch()


@onlyOnce
def patch_config_for_unit_tests():
    from buildbot import config
    # by default, buildbot.config warns about not configured buildbotNetUsageData.
    # its important for users to not leak information, but unneeded and painful for tests
    config._in_unit_tests = True


@onlyOnce
def patch_unittest_testcase():
    from twisted.trial.unittest import TestCase

    # In Python 3.2,
    # - assertRaisesRegexp() was renamed to assertRaisesRegex(),
    #   and assertRaisesRegexp() was deprecated.
    # - assertRegexpMatches() was renamed to assertRegex()
    #   and assertRegexpMatches() was deprecated.
    if not getattr(TestCase, "assertRaisesRegex", None):
        TestCase.assertRaisesRegex = TestCase.assertRaisesRegexp
    if not getattr(TestCase, "assertRegex", None):
        TestCase.assertRegex = TestCase.assertRegexpMatches


def patch_all(for_tests=False):
    if for_tests:
        patch_servicechecks()
        patch_testcase_timeout()
        patch_decorators()
        patch_mysqlclient_warnings()
        patch_config_for_unit_tests()
        patch_unittest_testcase()



# This file is part of Buildbot.  Buildbot is free software: you can
# redistribute it and/or modify it under the terms of the GNU General Public
# License as published by the Free Software Foundation, version 2.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
# details.
#
# You should have received a copy of the GNU General Public License along with
# this program; if not, write to the Free Software Foundation, Inc., 51
# Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#
# Copyright Buildbot Team Members


from twisted.internet import defer
from twisted.python import log

from buildbot.process.measured_service import MeasuredBuildbotServiceManager
from buildbot.util import misc
from buildbot.worker.protocols import pb as bbpb


class WorkerRegistration:

    __slots__ = ['master', 'worker', 'pbReg']

    def __init__(self, master, worker):
        self.master = master
        self.worker = worker

    def __repr__(self):
        return "<{} for {}>".format(self.__class__.__name__, repr(self.worker.workername))

    @defer.inlineCallbacks
    def unregister(self):
        bs = self.worker
        # update with portStr=None to remove any registration in place
        yield self.master.workers.pb.updateRegistration(
            bs.workername, bs.password, None)
        yield self.master.workers._unregister(self)

    @defer.inlineCallbacks
    def update(self, worker_config, global_config):
        # For most protocols, there's nothing to do, but for PB we must
        # update the registration in case the port or password has changed.
        if 'pb' in global_config.protocols:
            self.pbReg = yield self.master.workers.pb.updateRegistration(
                worker_config.workername, worker_config.password,
                global_config.protocols['pb']['port'])

    def getPBPort(self):
        return self.pbReg.getPort()


class WorkerManager(MeasuredBuildbotServiceManager):

    name = "WorkerManager"
    managed_services_name = "workers"

    config_attr = "workers"
    PING_TIMEOUT = 10
    reconfig_priority = 127

    def __init__(self, master):
        super().__init__()

        self.pb = bbpb.Listener()
        self.pb.setServiceParent(master)

        # WorkerRegistration instances keyed by worker name
        self.registrations = {}

        # connection objects keyed by worker name
        self.connections = {}

    @property
    def workers(self):
        # self.workers contains a ready Worker instance for each
        # potential worker, i.e. all the ones listed in the config file.
        # If the worker is connected, self.workers[workername].worker will
        # contain a RemoteReference to their Bot instance. If it is not
        # connected, that attribute will hold None.
        # workers attribute is actually just an alias to multiService's
        # namedService
        return self.namedServices

    def getWorkerByName(self, workerName):
        return self.registrations[workerName].worker

    def register(self, worker):
        # TODO: doc that reg.update must be called, too
        workerName = worker.workername
        reg = WorkerRegistration(self.master, worker)
        self.registrations[workerName] = reg
        return defer.succeed(reg)

    def _unregister(self, registration):
        del self.registrations[registration.worker.workername]

    @defer.inlineCallbacks
    def newConnection(self, conn, workerName):
        if workerName in self.connections:
            log.msg(("Got duplication connection from '{}'"
                     " starting arbitration procedure").format(workerName))
            old_conn = self.connections[workerName]
            try:
                yield misc.cancelAfter(self.PING_TIMEOUT,
                                       old_conn.remotePrint("master got a duplicate connection"))
                # if we get here then old connection is still alive, and new
                # should be rejected
                raise RuntimeError("rejecting duplicate worker")
            except defer.CancelledError:
                old_conn.loseConnection()
                log.msg("Connected worker '{}' ping timed out after {} seconds".format(workerName,
                        self.PING_TIMEOUT))
            except RuntimeError:
                raise
            except Exception as e:
                old_conn.loseConnection()
                log.msg("Got error while trying to ping connected worker {}:{}".format(workerName,
                                                                                       e))
            log.msg("Old connection for '{}' was lost, accepting new".format(workerName))

        try:
            yield conn.remotePrint(message="attached")
            info = yield conn.remoteGetWorkerInfo()
            log.msg("Got workerinfo from '{}'".format(workerName))
        except Exception as e:
            log.msg("Failed to communicate with worker '{}'\n{}".format(workerName, e))
            raise

        conn.info = info
        self.connections[workerName] = conn

        def remove():
            del self.connections[workerName]
        conn.notifyOnDisconnect(remove)

        # accept the connection
        return True



# This file is part of Buildbot.  Buildbot is free software: you can
# redistribute it and/or modify it under the terms of the GNU General Public
# License as published by the Free Software Foundation, version 2.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
# details.
#
# You should have received a copy of the GNU General Public License along with
# this program; if not, write to the Free Software Foundation, Inc., 51
# Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#
# Copyright Buildbot Team Members

from twisted.internet import defer

from buildbot.test.fakedb.base import FakeDBComponent
from buildbot.test.fakedb.row import Row


class Tag(Row):
    table = "tags"

    defaults = dict(
        id=None,
        name='some:tag',
        name_hash=None,
    )

    id_column = 'id'
    hashedColumns = [('name_hash', ('name',))]


class FakeTagsComponent(FakeDBComponent):

    def setUp(self):
        self.tags = {}

    def insertTestData(self, rows):
        for row in rows:
            if isinstance(row, Tag):
                self.tags[row.id] = dict(
                    id=row.id,
                    name=row.name)

    def findTagId(self, name):
        for m in self.tags.values():
            if m['name'] == name:
                return defer.succeed(m['id'])
        id = len(self.tags) + 1
        self.tags[id] = dict(
            id=id,
            name=name)
        return defer.succeed(id)



# This file is part of Buildbot. Buildbot is free software: you can
# redistribute it and/or modify it under the terms of the GNU General Public
# License as published by the Free Software Foundation, version 2.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE. See the GNU General Public License for more
# details.
#
# You should have received a copy of the GNU General Public License along with
# this program; if not, write to the Free Software Foundation, Inc., 51
# Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#
# Copyright Buildbot Team Members


import sqlalchemy as sa

from buildbot.util import sautils


def upgrade(migrate_engine):
    metadata = sa.MetaData()
    metadata.bind = migrate_engine

    sautils.Table(
        'builds', metadata,
        sa.Column('id', sa.Integer, primary_key=True),
        # ...
    )

    sautils.Table(
        'builders', metadata,
        sa.Column('id', sa.Integer, primary_key=True),
        # ...
    )

    sautils.Table(
        'steps', metadata,
        sa.Column('id', sa.Integer, primary_key=True),
        # ...
    )

    test_result_sets = sautils.Table(
        'test_result_sets', metadata,

        sa.Column('id', sa.Integer, primary_key=True),

        sa.Column('builderid', sa.Integer,
                  sa.ForeignKey('builders.id', ondelete='CASCADE'),
                  nullable=False),

        sa.Column('buildid', sa.Integer,
                  sa.ForeignKey('builds.id', ondelete='CASCADE'),
                  nullable=False),

        sa.Column('stepid', sa.Integer,
                  sa.ForeignKey('steps.id', ondelete='CASCADE'),
                  nullable=False),

        sa.Column('description', sa.Text, nullable=True),

        sa.Column('category', sa.Text, nullable=False),

        sa.Column('value_unit', sa.Text, nullable=False),

        sa.Column('tests_passed', sa.Integer, nullable=True),

        sa.Column('tests_failed', sa.Integer, nullable=True),

        sa.Column('complete', sa.SmallInteger, nullable=False),
    )

    test_results = sautils.Table(
        'test_results', metadata,

        sa.Column('id', sa.Integer, primary_key=True),

        sa.Column('builderid', sa.Integer,
                  sa.ForeignKey('builders.id', ondelete='CASCADE'),
                  nullable=False),

        sa.Column('test_result_setid', sa.Integer,
                  sa.ForeignKey('test_result_sets.id', ondelete='CASCADE'),
                  nullable=False),

        sa.Column('test_nameid', sa.Integer,
                  sa.ForeignKey('test_names.id', ondelete='CASCADE'),
                  nullable=True),

        sa.Column('test_code_pathid', sa.Integer,
                  sa.ForeignKey('test_code_paths.id', ondelete='CASCADE'),
                  nullable=True),

        sa.Column('line', sa.Integer, nullable=True),

        sa.Column('duration_ns', sa.Integer, nullable=True),

        sa.Column('value', sa.Text, nullable=False),
    )

    test_names = sautils.Table(
        'test_names', metadata,

        sa.Column('id', sa.Integer, primary_key=True),

        sa.Column('builderid', sa.Integer,
                  sa.ForeignKey('builders.id', ondelete='CASCADE'),
                  nullable=False),

        sa.Column('name', sa.Text, nullable=False),
    )

    test_code_paths = sautils.Table(
        'test_code_paths', metadata,

        sa.Column('id', sa.Integer, primary_key=True),

        sa.Column('builderid', sa.Integer,
                  sa.ForeignKey('builders.id', ondelete='CASCADE'),
                  nullable=False),

        sa.Column('path', sa.Text, nullable=False),
    )

    # create the tables
    test_result_sets.create()
    test_names.create()
    test_code_paths.create()
    test_results.create()

    # create indexes
    idx = sa.Index('test_names_name', test_names.c.builderid, test_names.c.name,
                   mysql_length={'name': 255})
    idx.create()

    idx = sa.Index('test_code_paths_path', test_code_paths.c.builderid, test_code_paths.c.path,
                   mysql_length={'path': 255})
    idx.create()



# This file is part of Buildbot.  Buildbot is free software: you can
# redistribute it and/or modify it under the terms of the GNU General Public
# License as published by the Free Software Foundation, version 2.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
# details.
#
# You should have received a copy of the GNU General Public License along with
# this program; if not, write to the Free Software Foundation, Inc., 51
# Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#
# Copyright Buildbot Team Members

from twisted.internet import defer
from zope.interface import implementer

from buildbot import interfaces
from buildbot.reporters import utils
from buildbot.reporters.message import MessageFormatter

from .utils import BuildStatusGeneratorMixin


@implementer(interfaces.IReportGenerator)
class BuildSetStatusGenerator(BuildStatusGeneratorMixin):

    wanted_event_keys = [
        ('buildsets', None, 'complete'),
    ]

    compare_attrs = ['formatter']

    def __init__(self, mode=("failing", "passing", "warnings"),
                 tags=None, builders=None, schedulers=None, branches=None,
                 subject="Buildbot %(result)s in %(title)s on %(builder)s",
                 add_logs=False, add_patch=False, message_formatter=None):
        super().__init__(mode, tags, builders, schedulers, branches, subject, add_logs, add_patch)
        self.formatter = message_formatter
        if self.formatter is None:
            self.formatter = MessageFormatter()

    @defer.inlineCallbacks
    def generate(self, master, reporter, key, message):
        bsid = message['bsid']
        res = yield utils.getDetailsForBuildset(master, bsid,
                                                wantProperties=self.formatter.wantProperties,
                                                wantSteps=self.formatter.wantSteps,
                                                wantPreviousBuild=self._want_previous_build(),
                                                wantLogs=self.formatter.wantLogs)

        builds = res['builds']
        buildset = res['buildset']

        # only include builds for which isMessageNeeded returns true
        builds = [build for build in builds
                  if self.is_message_needed_by_props(build) and
                  self.is_message_needed_by_results(build)]
        if not builds:
            return None

        report = yield self.build_message(self.formatter, master, reporter, "whole buildset",
                                          builds, buildset['results'])
        return report

    def _want_previous_build(self):
        return "change" in self.mode or "problem" in self.mode



# This file is part of Buildbot.  Buildbot is free software: you can
# redistribute it and/or modify it under the terms of the GNU General Public
# License as published by the Free Software Foundation, version 2.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
# details.
#
# You should have received a copy of the GNU General Public License along with
# this program; if not, write to the Free Software Foundation, Inc., 51
# Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#
# Copyright Buildbot Team Members

import sys

import twisted
from twisted.internet import defer
from twisted.internet import utils
from twisted.trial import reporter
from twisted.trial import unittest

from buildbot.test.util.gpo import Expect
from buildbot.test.util.gpo import GetProcessOutputMixin


class TestGPOMixin(unittest.TestCase):

    # these tests use self.patch, but the SkipTest exception gets eaten, so
    # explicitly skip things here.
    if twisted.version.major <= 9 and sys.version_info[:2] == (2, 7):
        skip = "unittest.TestCase.patch is not available"

    def runTestMethod(self, method):
        class TestCase(GetProcessOutputMixin, unittest.TestCase):

            def setUp(self):
                self.setUpGetProcessOutput()

            def runTest(self):
                return method(self)
        self.testcase = TestCase()
        result = reporter.TestResult()
        self.testcase.run(result)  # This blocks
        return result

    def assertTestFailure(self, result, expectedFailure):
        self.assertEqual(result.errors, [])
        self.assertEqual(len(result.failures), 1)
        self.assertTrue(result.failures[0][1].check(unittest.FailTest))
        if expectedFailure:
            self.assertSubstring(
                expectedFailure, result.failures[0][1].getErrorMessage())

    def assertSuccessful(self, result):
        if not result.wasSuccessful():
            output = 'expected success'
            if result.failures:
                output += ('\ntest failed: {}'.format(result.failures[0][1].getErrorMessage()))
            if result.errors:
                output += ('\nerrors: {}'.format([error[1].value for error in result.errors]))
            raise self.failureException(output)

        self.assertTrue(result.wasSuccessful())

    def test_patch(self):
        original_getProcessOutput = utils.getProcessOutput
        original_getProcessOutputAndValue = utils.getProcessOutputAndValue

        def method(testcase):
            testcase.expectCommands()
            self.assertEqual(utils.getProcessOutput,
                             testcase.patched_getProcessOutput)
            self.assertEqual(utils.getProcessOutputAndValue,
                             testcase.patched_getProcessOutputAndValue)
        result = self.runTestMethod(method)
        self.assertSuccessful(result)
        self.assertEqual(utils.getProcessOutput,
                         original_getProcessOutput)
        self.assertEqual(utils.getProcessOutputAndValue,
                         original_getProcessOutputAndValue)

    def test_methodChaining(self):
        expect = Expect('command')
        self.assertEqual(expect, expect.exit(0))
        self.assertEqual(expect, expect.stdout(b"output"))
        self.assertEqual(expect, expect.stderr(b"error"))

    def test_gpo_oneCommand(self):
        @defer.inlineCallbacks
        def method(testcase):
            testcase.expectCommands(Expect("command"))
            res = yield utils.getProcessOutput("command", ())
            self.assertEqual(res, b'')
            testcase.assertAllCommandsRan()
        result = self.runTestMethod(method)
        self.assertSuccessful(result)

    def test_gpo_expectTwo_runOne(self):
        @defer.inlineCallbacks
        def method(testcase):
            testcase.expectCommands(Expect("command"))
            testcase.expectCommands(Expect("command2"))
            res = yield utils.getProcessOutput("command", ())
            self.assertEqual(res, b'')
            testcase.assertAllCommandsRan()
        result = self.runTestMethod(method)
        self.assertTestFailure(result, "assert all expected commands were run")

    def test_gpo_wrongCommand(self):
        def method(testcase):
            testcase.expectCommands(Expect("command2"))
            d = utils.getProcessOutput("command", ())
            return d
        result = self.runTestMethod(method)
        self.assertTestFailure(result, "unexpected command run")
        # assert we have a meaningful message
        self.assertTestFailure(result, "command2")

    def test_gpo_wrongArgs(self):
        @defer.inlineCallbacks
        def method(testcase):
            testcase.expectCommands(Expect("command", "arg"))
            yield utils.getProcessOutput("command", ("otherarg",))
            testcase.assertAllCommandsRan()
        result = self.runTestMethod(method)
        self.assertTestFailure(result, "unexpected command run")

    def test_gpo_missingPath(self):
        @defer.inlineCallbacks
        def method(testcase):
            testcase.expectCommands(Expect("command", "arg").path("/home"))
            yield utils.getProcessOutput("command", ("otherarg",))
            testcase.assertAllCommandsRan()
        result = self.runTestMethod(method)
        self.assertTestFailure(result, "unexpected command run")

    def test_gpo_wrongPath(self):
        @defer.inlineCallbacks
        def method(testcase):
            testcase.expectCommands(Expect("command", "arg").path("/home"))
            yield utils.getProcessOutput("command", ("otherarg",), path="/work")
            testcase.assertAllCommandsRan()
        result = self.runTestMethod(method)
        self.assertTestFailure(result, "unexpected command run")

    def test_gpo_notCurrentPath(self):
        @defer.inlineCallbacks
        def method(testcase):
            testcase.expectCommands(Expect("command", "arg"))
            yield utils.getProcessOutput("command", ("otherarg",), path="/work")
            testcase.assertAllCommandsRan()
        result = self.runTestMethod(method)
        self.assertTestFailure(result, "unexpected command run")

    def test_gpo_errorOutput(self):
        def method(testcase):
            testcase.expectCommands(Expect("command").stderr(b"some test"))
            d = testcase.assertFailure(
                utils.getProcessOutput("command", ()), [IOError])
            return d
        result = self.runTestMethod(method)
        self.assertTestFailure(result, "got stderr: " + repr(b'some test'))

    def test_gpo_errorOutput_errtoo(self):
        @defer.inlineCallbacks
        def method(testcase):
            testcase.expectCommands(Expect("command").stderr(b"some test"))
            res = yield utils.getProcessOutput("command", (), errortoo=True)
            testcase.assertEqual(res, b"some test")
        result = self.runTestMethod(method)
        self.assertSuccessful(result)

    def test_gpo_exitIgnored(self):
        @defer.inlineCallbacks
        def method(testcase):
            testcase.expectCommands(Expect("command").exit(1))
            res = yield utils.getProcessOutput("command", ())
            self.assertEqual(res, b'')
        result = self.runTestMethod(method)
        self.assertSuccessful(result)

    def test_gpo_output(self):
        @defer.inlineCallbacks
        def method(testcase):
            testcase.expectCommands(Expect("command").stdout(b"stdout"))
            res = yield utils.getProcessOutput("command", ())
            testcase.assertEqual(res, b"stdout")
        result = self.runTestMethod(method)
        self.assertSuccessful(result)

    def test_gpo_outputAndError(self):
        @defer.inlineCallbacks
        def method(testcase):
            testcase.expectCommands(
                Expect("command").stdout(b"stdout").stderr(b"stderr"))
            res = yield utils.getProcessOutput("command", (), errortoo=True)

            testcase.assertSubstring(b"stdout", res)
            testcase.assertSubstring(b"stderr", res)
        result = self.runTestMethod(method)
        self.assertSuccessful(result)

    def test_gpo_environ_success(self):
        @defer.inlineCallbacks
        def method(testcase):
            testcase.expectCommands(Expect("command"))
            testcase.addGetProcessOutputExpectEnv({'key': 'value'})
            res = yield utils.getProcessOutput("command", (), env={'key': 'value'})
            self.assertEqual(res, b'')
            testcase.assertAllCommandsRan()
        result = self.runTestMethod(method)
        self.assertSuccessful(result)

    def test_gpo_environ_wrongValue(self):
        def method(testcase):
            testcase.expectCommands(Expect("command"))
            testcase.addGetProcessOutputExpectEnv({'key': 'value'})
            d = utils.getProcessOutput(
                "command", (), env={'key': 'wrongvalue'})
            return d
        result = self.runTestMethod(method)
        self.assertTestFailure(
            result, "Expected environment to have key = 'value'")

    def test_gpo_environ_missing(self):
        def method(testcase):
            testcase.expectCommands(Expect("command"))
            testcase.addGetProcessOutputExpectEnv({'key': 'value'})
            d = utils.getProcessOutput("command", ())
            return d
        result = self.runTestMethod(method)
        self.assertTestFailure(
            result, "Expected environment to have key = 'value'")

    def test_gpoav_oneCommand(self):
        @defer.inlineCallbacks
        def method(testcase):
            testcase.expectCommands(Expect("command"))
            res = yield utils.getProcessOutputAndValue("command", ())
            self.assertEqual(res, (b'', b'', 0))
            testcase.assertAllCommandsRan()
        result = self.runTestMethod(method)
        self.assertSuccessful(result)

    def test_gpoav_expectTwo_runOne(self):
        @defer.inlineCallbacks
        def method(testcase):
            testcase.expectCommands(Expect("command"))
            testcase.expectCommands(Expect("command2"))
            res = yield utils.getProcessOutputAndValue("command", ())
            self.assertEqual(res, (b'', b'', 0))
            testcase.assertAllCommandsRan()
        result = self.runTestMethod(method)
        self.assertTestFailure(result, "assert all expected commands were run")

    def test_gpoav_wrongCommand(self):
        @defer.inlineCallbacks
        def method(testcase):
            testcase.expectCommands(Expect("command2"))
            yield utils.getProcessOutputAndValue("command", ())
            testcase.assertAllCommandsRan()
        result = self.runTestMethod(method)
        self.assertTestFailure(result, "unexpected command run")

    def test_gpoav_wrongArgs(self):
        @defer.inlineCallbacks
        def method(testcase):
            testcase.expectCommands(Expect("command", "arg"))
            yield utils.getProcessOutputAndValue("command", ("otherarg",))
            testcase.assertAllCommandsRan()
        result = self.runTestMethod(method)
        self.assertTestFailure(result, "unexpected command run")

    def test_gpoav_missingPath(self):
        @defer.inlineCallbacks
        def method(testcase):
            testcase.expectCommands(Expect("command", "arg").path("/home"))
            yield utils.getProcessOutputAndValue("command", ("otherarg",))
            testcase.assertAllCommandsRan()
        result = self.runTestMethod(method)
        self.assertTestFailure(result, "unexpected command run")

    def test_gpoav_wrongPath(self):
        @defer.inlineCallbacks
        def method(testcase):
            testcase.expectCommands(Expect("command", "arg").path("/home"))
            yield utils.getProcessOutputAndValue(
                    "command", ("otherarg",), path="/work")
            testcase.assertAllCommandsRan()
        result = self.runTestMethod(method)
        self.assertTestFailure(result, "unexpected command run")

    def test_gpoav_notCurrentPath(self):
        @defer.inlineCallbacks
        def method(testcase):
            testcase.expectCommands(Expect("command", "arg"))
            yield utils.getProcessOutputAndValue(
                    "command", ("otherarg",), path="/work")
            testcase.assertAllCommandsRan()
        result = self.runTestMethod(method)
        self.assertTestFailure(result, "unexpected command run")

    def test_gpoav_errorOutput(self):
        @defer.inlineCallbacks
        def method(testcase):
            testcase.expectCommands(Expect("command").stderr(b"some test"))
            res = yield utils.getProcessOutputAndValue("command", ())
            self.assertEqual(res, (b'', b'some test', 0))
        result = self.runTestMethod(method)
        self.assertSuccessful(result)

    def test_gpoav_exit(self):
        @defer.inlineCallbacks
        def method(testcase):
            testcase.expectCommands(Expect("command").exit(1))
            res = yield utils.getProcessOutputAndValue("command", ())
            self.assertEqual(res, (b'', b'', 1))
        result = self.runTestMethod(method)
        self.assertSuccessful(result)

    def test_gpoav_output(self):
        @defer.inlineCallbacks
        def method(testcase):
            testcase.expectCommands(Expect("command").stdout(b"stdout"))
            res = yield utils.getProcessOutputAndValue("command", ())
            testcase.assertEqual(res, (b"stdout", b'', 0))
        result = self.runTestMethod(method)
        self.assertSuccessful(result)

    def test_gpoav_outputAndError(self):
        @defer.inlineCallbacks
        def method(testcase):
            testcase.expectCommands(
                Expect("command").stdout(b"stdout").stderr(b"stderr"))
            res = yield utils.getProcessOutputAndValue("command", ())
            testcase.assertEqual(res, (b"stdout", b'stderr', 0))

        result = self.runTestMethod(method)
        self.assertSuccessful(result)



# This file is part of Buildbot.  Buildbot is free software: you can
# redistribute it and/or modify it under the terms of the GNU General Public
# License as published by the Free Software Foundation, version 2.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
# details.
#
# You should have received a copy of the GNU General Public License along with
# this program; if not, write to the Free Software Foundation, Inc., 51
# Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#
# Copyright Buildbot Team Members


from twisted.internet import defer

from buildbot.process.results import FAILURE
from buildbot.process.results import SUCCESS
from buildbot.steps.source.base import Source
from buildbot.test.util.decorators import skipUnlessPlatformIs
from buildbot.test.util.integration import RunMasterBase

# a simple patch which adds a Makefile
PATCH = """diff --git a/Makefile b/Makefile
new file mode 100644
index 0000000..8a5cf80
--- /dev/null
+++ b/Makefile
@@ -0,0 +1,2 @@
+all:
+\techo OK
"""


class MySource(Source):
    """A source class which only applies the patch"""

    @defer.inlineCallbacks
    def run_vc(self, branch, revision, patch):
        self.stdio_log = yield self.addLogForRemoteCommands("stdio")
        if patch:
            yield self.patch(patch)
        return SUCCESS


class ShellMaster(RunMasterBase):

    @skipUnlessPlatformIs("posix")  # make is not installed on windows
    @defer.inlineCallbacks
    def test_shell(self):
        yield self.setupConfig(masterConfig())
        build = yield self.doForceBuild(wantSteps=True, wantLogs=True,
                                        forceParams={'foo_patch_body': PATCH})
        self.assertEqual(build['buildid'], 1)
        # if makefile was not properly created, we would have a failure
        self.assertEqual(build['results'], SUCCESS)

    @defer.inlineCallbacks
    def test_shell_no_patch(self):
        yield self.setupConfig(masterConfig())
        build = yield self.doForceBuild(wantSteps=True, wantLogs=True)
        self.assertEqual(build['buildid'], 1)
        # if no patch, the source step is happy, but the make step cannot find makefile
        self.assertEqual(build['steps'][1]['results'], SUCCESS)
        self.assertEqual(build['steps'][2]['results'], FAILURE)
        self.assertEqual(build['results'], FAILURE)


# master configuration
def masterConfig():
    c = {}
    from buildbot.config import BuilderConfig
    from buildbot.process.factory import BuildFactory
    from buildbot.plugins import steps, schedulers, util

    c['schedulers'] = [
        schedulers.ForceScheduler(
            name="force",
            codebases=[util.CodebaseParameter(
                "foo", patch=util.PatchParameter())],
            builderNames=["testy"])]

    f = BuildFactory()
    f.addStep(MySource(codebase='foo'))
    # if the patch was applied correctly, then make will work!
    f.addStep(steps.ShellCommand(command=["make"]))
    c['builders'] = [
        BuilderConfig(name="testy",
                      workernames=["local1"],
                      factory=f)]
    return c



# This file is part of Buildbot.  Buildbot is free software: you can
# redistribute it and/or modify it under the terms of the GNU General Public
# License as published by the Free Software Foundation, version 2.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
# details.
#
# You should have received a copy of the GNU General Public License along with
# this program; if not, write to the Free Software Foundation, Inc., 51
# Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#
# Copyright Buildbot Team Members


from twisted.internet import defer

from buildbot.plugins import steps
from buildbot.process.results import EXCEPTION
from buildbot.process.results import SUCCESS
from buildbot.test.util.integration import RunMasterBase


class TestLog(RunMasterBase):
    # master configuration

    def masterConfig(self, step):
        c = {}
        from buildbot.config import BuilderConfig
        from buildbot.process.factory import BuildFactory
        from buildbot.plugins import schedulers

        c['schedulers'] = [
            schedulers.AnyBranchScheduler(
                name="sched",
                builderNames=["testy"])]

        f = BuildFactory()
        f.addStep(step)
        c['builders'] = [
            BuilderConfig(name="testy",
                          workernames=["local1"],
                          factory=f)]
        return c

    @defer.inlineCallbacks
    def test_shellcommand(self):

        class MyStep(steps.ShellCommandNewStyle):

            def _newLog(obj, name, type, logid, logEncoding):
                r = steps.ShellCommand._newLog(obj, name, type, logid, logEncoding)
                self.curr_log = r
                return self.curr_log

        step = MyStep(command='echo hello')

        yield self.setupConfig(self.masterConfig(step))

        change = dict(branch="master",
                      files=["foo.c"],
                      author="me@foo.com",
                      committer="me@foo.com",
                      comments="good stuff",
                      revision="HEAD",
                      project="none")
        build = yield self.doForceBuild(wantSteps=True, useChange=change, wantLogs=True)
        self.assertEqual(build['buildid'], 1)
        self.assertEqual(build['results'], SUCCESS)
        self.assertTrue(self.curr_log.finished)

    @defer.inlineCallbacks
    def test_mastershellcommand(self):

        class MyStep(steps.MasterShellCommandNewStyle):

            def _newLog(obj, name, type, logid, logEncoding):
                r = steps.MasterShellCommandNewStyle._newLog(obj, name, type, logid, logEncoding)
                self.curr_log = r
                return self.curr_log

        step = MyStep(command='echo hello')

        yield self.setupConfig(self.masterConfig(step))

        change = dict(branch="master",
                      files=["foo.c"],
                      author="me@foo.com",
                      committer="me@foo.com",
                      comments="good stuff",
                      revision="HEAD",
                      project="none")
        build = yield self.doForceBuild(wantSteps=True, useChange=change, wantLogs=True)
        self.assertEqual(build['buildid'], 1)
        self.assertEqual(build['results'], SUCCESS)
        self.assertTrue(self.curr_log.finished)

    @defer.inlineCallbacks
    def test_mastershellcommand_issue(self):

        class MyStep(steps.MasterShellCommandNewStyle):

            def _newLog(obj, name, type, logid, logEncoding):
                r = steps.MasterShellCommandNewStyle._newLog(obj, name, type, logid, logEncoding)
                self.curr_log = r
                self.patch(r, "finish", lambda: defer.fail(RuntimeError('Could not finish')))
                return self.curr_log

        step = MyStep(command='echo hello')

        yield self.setupConfig(self.masterConfig(step))

        change = dict(branch="master",
                      files=["foo.c"],
                      author="me@foo.com",
                      committer="me@foo.com",
                      comments="good stuff",
                      revision="HEAD",
                      project="none")
        build = yield self.doForceBuild(wantSteps=True, useChange=change, wantLogs=True)
        self.assertEqual(build['buildid'], 1)
        self.assertFalse(self.curr_log.finished)
        self.assertEqual(build['results'], EXCEPTION)
        errors = self.flushLoggedErrors()
        self.assertEqual(len(errors), 1)
        error = errors[0]
        self.assertEqual(error.getErrorMessage(), 'Could not finish')



# This file is part of Buildbot.  Buildbot is free software: you can
# redistribute it and/or modify it under the terms of the GNU General Public
# License as published by the Free Software Foundation, version 2.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
# details.
#
# You should have received a copy of the GNU General Public License along with
# this program; if not, write to the Free Software Foundation, Inc., 51
# Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#
# Portions Copyright Buildbot Team Members
# Portions Copyright 2010 Isotoma Limited


import os

from twisted.internet import defer
from twisted.internet import threads
from twisted.internet import utils
from twisted.python import failure
from twisted.python import log

from buildbot import config
from buildbot.util.eventual import eventually
from buildbot.worker import AbstractLatentWorker

try:
    import libvirt
except ImportError:
    libvirt = None


class WorkQueue:

    """
    I am a class that turns parallel access into serial access.

    I exist because we want to run libvirt access in threads as we don't
    trust calls not to block, but under load libvirt doesn't seem to like
    this kind of threaded use.
    """

    def __init__(self):
        self.queue = []

    def _process(self):
        log.msg("Looking to start a piece of work now...")

        # Is there anything to do?
        if not self.queue:
            log.msg("_process called when there is no work")
            return

        # Peek at the top of the stack - get a function to call and
        # a deferred to fire when its all over
        d, next_operation, args, kwargs = self.queue[0]

        # Start doing some work - expects a deferred
        try:
            d2 = next_operation(*args, **kwargs)
        except Exception:
            d2 = defer.fail()

        # Whenever a piece of work is done, whether it worked or not
        # call this to schedule the next piece of work
        @d2.addBoth
        def _work_done(res):
            log.msg("Completed a piece of work")
            self.queue.pop(0)
            if self.queue:
                log.msg("Preparing next piece of work")
                eventually(self._process)
            return res

        # When the work is done, trigger d
        d2.chainDeferred(d)

    def execute(self, cb, *args, **kwargs):
        kickstart_processing = not self.queue
        d = defer.Deferred()
        self.queue.append((d, cb, args, kwargs))
        if kickstart_processing:
            self._process()
        return d

    def executeInThread(self, cb, *args, **kwargs):
        return self.execute(threads.deferToThread, cb, *args, **kwargs)


# A module is effectively a singleton class, so this is OK
queue = WorkQueue()


class Domain:

    """
    I am a wrapper around a libvirt Domain object
    """

    def __init__(self, connection, domain):
        self.connection = connection
        self.domain = domain

    def name(self):
        return queue.executeInThread(self.domain.name)

    def create(self):
        return queue.executeInThread(self.domain.create)

    def shutdown(self):
        return queue.executeInThread(self.domain.shutdown)

    def destroy(self):
        return queue.executeInThread(self.domain.destroy)


class Connection:

    """
    I am a wrapper around a libvirt Connection object.
    """

    DomainClass = Domain

    def __init__(self, uri):
        self.uri = uri
        self.connection = libvirt.open(uri)

    @defer.inlineCallbacks
    def lookupByName(self, name):
        """ I lookup an existing predefined domain """
        res = yield queue.executeInThread(self.connection.lookupByName, name)
        return self.DomainClass(self, res)

    @defer.inlineCallbacks
    def create(self, xml):
        """ I take libvirt XML and start a new VM """
        res = yield queue.executeInThread(self.connection.createXML, xml, 0)
        return self.DomainClass(self, res)

    @defer.inlineCallbacks
    def all(self):
        domains = []
        domain_ids = yield queue.executeInThread(self.connection.listDomainsID)

        for did in domain_ids:
            domain = yield queue.executeInThread(self.connection.lookupByID, did)
            domains.append(self.DomainClass(self, domain))

        return domains


class LibVirtWorker(AbstractLatentWorker):

    def __init__(self, name, password, connection, hd_image, base_image=None, xml=None,
                 **kwargs):
        super().__init__(name, password, **kwargs)
        if not libvirt:
            config.error(
                "The python module 'libvirt' is needed to use a LibVirtWorker")

        self.connection = connection
        self.image = hd_image
        self.base_image = base_image
        self.xml = xml

        self.cheap_copy = True
        self.graceful_shutdown = False

        self.domain = None

        self.ready = False
        self._find_existing_deferred = self._find_existing_instance()

    @defer.inlineCallbacks
    def _find_existing_instance(self):
        """
        I find existing VMs that are already running that might be orphaned instances of this
        worker.
        """
        if not self.connection:
            return None

        domains = yield self.connection.all()
        for d in domains:
            name = yield d.name()
            if name.startswith(self.workername):
                self.domain = d
                break

        self.ready = True
        return None

    def canStartBuild(self):
        if not self.ready:
            log.msg("Not accepting builds as existing domains not iterated")
            return False

        if self.domain and not self.isConnected():
            log.msg(
                "Not accepting builds as existing domain but worker not connected")
            return False

        return super().canStartBuild()

    def _prepare_base_image(self):
        """
        I am a private method for creating (possibly cheap) copies of a
        base_image for start_instance to boot.
        """
        if not self.base_image:
            return defer.succeed(True)

        if self.cheap_copy:
            clone_cmd = "qemu-img"
            clone_args = "create -b %(base)s -f qcow2 %(image)s"
        else:
            clone_cmd = "cp"
            clone_args = "%(base)s %(image)s"

        clone_args = clone_args % {
            "base": self.base_image,
            "image": self.image,
        }

        log.msg("Cloning base image: {} {}'".format(clone_cmd, clone_args))

        d = utils.getProcessValue(clone_cmd, clone_args.split())

        def _log_result(res):
            log.msg("Cloning exit code was: %d" % res)
            return res

        def _log_error(err):
            log.err("Cloning failed: {}".format(err))
            return err

        d.addCallbacks(_log_result, _log_error)

        return d

    @defer.inlineCallbacks
    def start_instance(self, build):
        """
        I start a new instance of a VM.

        If a base_image is specified, I will make a clone of that otherwise i will
        use image directly.

        If i'm not given libvirt domain definition XML, I will look for my name
        in the list of defined virtual machines and start that.
        """
        if self.domain is not None:
            log.msg("Cannot start_instance '{}' as already active".format(self.workername))
            return False

        yield self._prepare_base_image()

        try:
            if self.xml:
                self.domain = yield self.connection.create(self.xml)
            else:
                self.domain = yield self.connection.lookupByName(self.workername)
                yield self.domain.create()
        except Exception:
            log.err(failure.Failure(),
                    ("Cannot start a VM ({}), failing gracefully and triggering"
                     "a new build check").format(self.workername))
            self.domain = None
            return False

        return True

    def stop_instance(self, fast=False):
        """
        I attempt to stop a running VM.
        I make sure any connection to the worker is removed.
        If the VM was using a cloned image, I remove the clone
        When everything is tidied up, I ask that bbot looks for work to do
        """
        @defer.inlineCallbacks
        def _destroy_domain(res, domain):
            log.msg('Graceful shutdown failed. Force destroying domain {}'.format(self.workername))
            # Don't return res to stop propagating shutdown error if destroy
            # was successful.
            yield domain.destroy()

        log.msg("Attempting to stop '{}'".format(self.workername))
        if self.domain is None:
            log.msg("I don't think that domain is even running, aborting")
            return defer.succeed(None)

        domain = self.domain
        self.domain = None

        if self.graceful_shutdown and not fast:
            log.msg("Graceful shutdown chosen for {}".format(self.workername))
            d = domain.shutdown()
            d.addErrback(_destroy_domain, domain)
        else:
            d = domain.destroy()

        if self.base_image:
            @d.addBoth
            def _remove_image(res):
                log.msg('Removing base image {} for {}'.format(self.image, self.workername))
                os.remove(self.image)
                return res

        return d



# This file is part of Buildbot.  Buildbot is free software: you can
# redistribute it and/or modify it under the terms of the GNU General Public
# License as published by the Free Software Foundation, version 2.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
# details.
#
# You should have received a copy of the GNU General Public License along with
# this program; if not, write to the Free Software Foundation, Inc., 51
# Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#
# Copyright Buildbot Team Members

import sqlalchemy as sa

from twisted.internet import defer

from buildbot.db import base
from buildbot.util import identifiers


class WorkersConnectorComponent(base.DBConnectorComponent):
    # Documentation is in developer/database.rst

    def findWorkerId(self, name):
        tbl = self.db.model.workers
        # callers should verify this and give good user error messages
        assert identifiers.isIdentifier(50, name)
        return self.findSomethingId(
            tbl=tbl,
            whereclause=(tbl.c.name == name),
            insert_values=dict(
                name=name,
                info={},
                paused=0,
                graceful=0,
            ))

    def _deleteFromConfiguredWorkers_thd(self, conn, buildermasterids, workerid=None):
        cfg_tbl = self.db.model.configured_workers
        # batch deletes to avoid using too many variables
        for batch in self.doBatch(buildermasterids, 100):
            q = cfg_tbl.delete()
            q = q.where(cfg_tbl.c.buildermasterid.in_(batch))
            if workerid:
                q = q.where(cfg_tbl.c.workerid == workerid)
            conn.execute(q).close()

    # returns a Deferred which returns None
    def deconfigureAllWorkersForMaster(self, masterid):
        def thd(conn):
            # first remove the old configured buildermasterids for this master and worker
            # as sqlalchemy does not support delete with join, we need to do
            # that in 2 queries
            cfg_tbl = self.db.model.configured_workers
            bm_tbl = self.db.model.builder_masters
            j = cfg_tbl
            j = j.outerjoin(bm_tbl)
            q = sa.select(
                [cfg_tbl.c.buildermasterid], from_obj=[j], distinct=True)
            q = q.where(bm_tbl.c.masterid == masterid)
            res = conn.execute(q)
            buildermasterids = [row['buildermasterid']
                                for row in res]
            res.close()
            self._deleteFromConfiguredWorkers_thd(conn, buildermasterids)

        return self.db.pool.do(thd)

    # returns a Deferred that returns None
    def workerConfigured(self, workerid, masterid, builderids):

        def thd(conn):

            cfg_tbl = self.db.model.configured_workers
            bm_tbl = self.db.model.builder_masters

            # get the buildermasterids that are configured
            if builderids:
                q = sa.select([bm_tbl.c.id], from_obj=[bm_tbl])
                q = q.where(bm_tbl.c.masterid == masterid)
                q = q.where(bm_tbl.c.builderid.in_(builderids))
                res = conn.execute(q)
                buildermasterids = {row['id'] for row in res}
                res.close()
            else:
                buildermasterids = set([])

            j = cfg_tbl
            j = j.outerjoin(bm_tbl)
            q = sa.select(
                [cfg_tbl.c.buildermasterid], from_obj=[j], distinct=True)
            q = q.where(bm_tbl.c.masterid == masterid)
            q = q.where(cfg_tbl.c.workerid == workerid)
            res = conn.execute(q)
            oldbuildermasterids = {row['buildermasterid'] for row in res}
            res.close()

            todeletebuildermasterids = oldbuildermasterids - buildermasterids
            toinsertbuildermasterids = buildermasterids - oldbuildermasterids
            transaction = conn.begin()
            self._deleteFromConfiguredWorkers_thd(conn, todeletebuildermasterids, workerid)

            # and insert the new ones
            if toinsertbuildermasterids:
                q = cfg_tbl.insert()
                conn.execute(q,
                             [{'workerid': workerid, 'buildermasterid': buildermasterid}
                              for buildermasterid in toinsertbuildermasterids]).close()

            transaction.commit()

        return self.db.pool.do(thd)

    @defer.inlineCallbacks
    def getWorker(self, workerid=None, name=None, masterid=None,
                  builderid=None):
        if workerid is None and name is None:
            return None
        workers = yield self.getWorkers(_workerid=workerid,
                                        _name=name, masterid=masterid, builderid=builderid)
        if workers:
            return workers[0]
        return None

    # returns a Deferred that returns a value
    def getWorkers(self, _workerid=None, _name=None, masterid=None,
                   builderid=None, paused=None, graceful=None):
        def thd(conn):
            workers_tbl = self.db.model.workers
            conn_tbl = self.db.model.connected_workers
            cfg_tbl = self.db.model.configured_workers
            bm_tbl = self.db.model.builder_masters

            def selectWorker(q):
                return q

            # first, get the worker itself and the configured_on info
            j = workers_tbl
            j = j.outerjoin(cfg_tbl)
            j = j.outerjoin(bm_tbl)
            q = sa.select(
                [workers_tbl.c.id, workers_tbl.c.name, workers_tbl.c.info,
                 workers_tbl.c.paused, workers_tbl.c.graceful,
                 bm_tbl.c.builderid, bm_tbl.c.masterid],
                from_obj=[j],
                order_by=[workers_tbl.c.id])

            if _workerid is not None:
                q = q.where(workers_tbl.c.id == _workerid)
            if _name is not None:
                q = q.where(workers_tbl.c.name == _name)
            if masterid is not None:
                q = q.where(bm_tbl.c.masterid == masterid)
            if builderid is not None:
                q = q.where(bm_tbl.c.builderid == builderid)
            if paused is not None:
                q = q.where(workers_tbl.c.paused == int(paused))
            if graceful is not None:
                q = q.where(workers_tbl.c.graceful == int(graceful))

            rv = {}
            res = None
            lastId = None
            cfgs = None
            for row in conn.execute(q):
                if row.id != lastId:
                    lastId = row.id
                    cfgs = []
                    res = {
                        'id': lastId,
                        'name': row.name,
                        'configured_on': cfgs,
                        'connected_to': [],
                        'workerinfo': row.info,
                        'paused': bool(row.paused),
                        'graceful': bool(row.graceful)}
                    rv[lastId] = res
                if row.builderid and row.masterid:
                    cfgs.append({'builderid': row.builderid,
                                 'masterid': row.masterid})

            # now go back and get the connection info for the same set of
            # workers
            j = conn_tbl
            if _name is not None:
                # note this is not an outer join; if there are unconnected
                # workers, they were captured in rv above
                j = j.join(workers_tbl)
            q = sa.select(
                [conn_tbl.c.workerid, conn_tbl.c.masterid],
                from_obj=[j],
                order_by=[conn_tbl.c.workerid])

            if _workerid is not None:
                q = q.where(conn_tbl.c.workerid == _workerid)
            if _name is not None:
                q = q.where(workers_tbl.c.name == _name)
            if masterid is not None:
                q = q.where(conn_tbl.c.masterid == masterid)

            for row in conn.execute(q):
                id = row.workerid
                if id not in rv:
                    continue
                rv[row.workerid]['connected_to'].append(row.masterid)

            return list(rv.values())
        return self.db.pool.do(thd)

    # returns a Deferred that returns None
    def workerConnected(self, workerid, masterid, workerinfo):
        def thd(conn):
            conn_tbl = self.db.model.connected_workers
            q = conn_tbl.insert()
            try:
                conn.execute(q,
                             {'workerid': workerid, 'masterid': masterid})
            except (sa.exc.IntegrityError, sa.exc.ProgrammingError):
                # if the row is already present, silently fail..
                pass

            bs_tbl = self.db.model.workers
            q = bs_tbl.update(whereclause=(bs_tbl.c.id == workerid))
            conn.execute(q, info=workerinfo)
        return self.db.pool.do(thd)

    # returns a Deferred that returns None
    def workerDisconnected(self, workerid, masterid):
        def thd(conn):
            tbl = self.db.model.connected_workers
            q = tbl.delete(whereclause=(tbl.c.workerid == workerid) &
                                       (tbl.c.masterid == masterid))
            conn.execute(q)
        return self.db.pool.do(thd)

    # returns a Deferred that returns None
    def setWorkerState(self, workerid, paused, graceful):
        def thd(conn):
            tbl = self.db.model.workers
            q = tbl.update(whereclause=(tbl.c.id == workerid))
            conn.execute(q, paused=int(paused), graceful=int(graceful))
        return self.db.pool.do(thd)



# This file is part of Buildbot.  Buildbot is free software: you can
# redistribute it and/or modify it under the terms of the GNU General Public
# License as published by the Free Software Foundation, version 2.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
# details.
#
# You should have received a copy of the GNU General Public License along with
# this program; if not, write to the Free Software Foundation, Inc., 51
# Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#
# Copyright Buildbot Team Members


from twisted.internet import defer
from twisted.internet import task

from buildbot.test.fake import fakemaster


class ChangeSourceMixin:

    """
    This class is used for testing change sources, and handles a few things:

     - starting and stopping a ChangeSource service
     - a fake master with a data API implementation
    """

    changesource = None
    started = False

    DUMMY_CHANGESOURCE_ID = 20
    OTHER_MASTER_ID = 93
    DEFAULT_NAME = "ChangeSource"

    def setUpChangeSource(self):
        "Set up the mixin - returns a deferred."
        self.master = fakemaster.make_master(self, wantDb=True, wantData=True)
        assert not hasattr(self.master, 'addChange')  # just checking..
        return defer.succeed(None)

    @defer.inlineCallbacks
    def tearDownChangeSource(self):
        "Tear down the mixin - returns a deferred."
        if not self.started:
            return
        if self.changesource.running:
            yield self.changesource.stopService()
        yield self.changesource.disownServiceParent()
        return

    def attachChangeSource(self, cs):
        "Set up a change source for testing; sets its .master attribute"
        self.changesource = cs
        # FIXME some changesource does not have master property yet but
        # mailchangesource has :-/
        try:
            self.changesource.master = self.master
        except AttributeError:
            self.changesource.setServiceParent(self.master)

        # configure the service to let secret manager render the secrets
        d = self.changesource.configureService()
        d.addErrback(lambda _: None)
        # also, now that changesources are ClusteredServices, setting up
        # the clock here helps in the unit tests that check that behavior
        self.changesource.clock = task.Clock()

    def startChangeSource(self):
        "start the change source as a service"
        self.started = True
        return self.changesource.startService()

    @defer.inlineCallbacks
    def stopChangeSource(self):
        "stop the change source again; returns a deferred"
        yield self.changesource.stopService()

        self.started = False

    def setChangeSourceToMaster(self, otherMaster):
        # some tests build the CS late, so for those tests we will require that
        # they use the default name in order to run tests that require master
        # assignments
        if self.changesource is not None:
            name = self.changesource.name
        else:
            name = self.DEFAULT_NAME

        self.master.data.updates.changesourceIds[
            name] = self.DUMMY_CHANGESOURCE_ID
        if otherMaster:
            self.master.data.updates.changesourceMasters[
                self.DUMMY_CHANGESOURCE_ID] = otherMaster
        else:
            del self.master.data.updates.changesourceMasters[
                self.DUMMY_CHANGESOURCE_ID]
